# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.


def order_check(chart_info: dict, ground_truth: dict, sort_by: str):
    order = ground_truth["sort"]
    encoding = chart_info["encoding"]
    data = chart_info["data"]
    channel_map = chart_info["channel_map"]

    if order is not None:
        # bar, line
        if sort_by == "axis":
            order_channel = order["channel"]
        else:
            order_channel = channel_map[order["channel"]]

        other_channel = "y" if order_channel == "x" else "x"

        order_channel_scale = encoding[order_channel]["scale"]
        other_channel_scale = encoding[other_channel]["scale"]

        # origin channel
        if (
            channel_map[order_channel] == "x"
            or channel_map[order_channel] == "classify"
        ):
            arr = []
            if len(order_channel_scale["range"]) == 0:
                scale_range = range(1, 1 + len(order_channel_scale["domain"]))
            else:
                scale_range = order_channel_scale["range"]

            for index in range(len(order_channel_scale["domain"])):
                arr.append(
                    tuple(
                        [
                            order_channel_scale["domain"][index],
                            scale_range[index],
                        ]
                    )
                )
            if order_channel == "x":
                reverse = True
            else:
                reverse = False
            if order["order"] == "ascending":
                arr.sort(key=lambda x: x[0], reverse=reverse)
            elif order["order"] == "descending":
                arr.sort(key=lambda x: x[0], reverse=not reverse)
            else:  # custom order
                sort_order = {}
                for index in range(len(order["order"])):
                    sort_order[order["order"][index]] = index
                arr.sort(key=lambda x: sort_order[x[0]], reverse=reverse)

            is_sorted = all([arr[i][1] > arr[i + 1][1] for i in range(len(arr) - 1)])
        # 'quantitative'
        else:
            # sort by other channel
            values_other = []
            if (
                "type" not in other_channel_scale
                or other_channel_scale["type"] == "ordinal"
            ):
                for index in range(len(other_channel_scale["domain"])):
                    values_other.append(
                        tuple(
                            [
                                other_channel_scale["domain"][index],
                                other_channel_scale["range"][index],
                            ]
                        )
                    )
                values_other.sort(key=lambda x: x[1])
                values_other = [item[0] for item in values_other]
            else:
                values_other = list(
                    set([datum["field_" + other_channel] for datum in data])
                )
                values_other.sort(
                    reverse=True
                    if (
                        other_channel_scale["domain"][1]
                        - other_channel_scale["domain"][0]
                    )
                    / (
                        other_channel_scale["range"][1]
                        - other_channel_scale["range"][0]
                    )
                    < 0
                    else False
                )

            # cumulative
            values_order = []
            for value in values_other:
                data_filter = [
                    float(d["field_" + order_channel])
                    for d in data
                    if d["field_" + other_channel] == value
                ]
                values_order.append(sum(data_filter))

            # filter zero data
            if chart_info["mark"] == "bar":
                values_order = list(filter(lambda x: x != 0, values_order))

            is_sorted = True
            if order["order"] == "ascending":
                is_sorted = all(
                    [
                        values_order[i] <= values_order[i + 1]
                        for i in range(len(values_order) - 1)
                    ]
                )
            elif order["order"] == "descending":
                is_sorted = all(
                    [
                        values_order[i] >= values_order[i + 1]
                        for i in range(len(values_order) - 1)
                    ]
                )

        if not is_sorted:
            return False, "Doesn't sort."
        else:
            return True, "Sorted."
    else:
        return True, "No sort."



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import json
from pathlib import Path

import pytest

from viseval.check import chart_check, deconstruct

folder = Path(__file__).resolve().parent

with open(folder / "assets/samples.json") as f:
    benchmark = json.load(f)


def test_chart_check_pie_4():
    with open(folder / "assets/pie_4_0.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        ground_truth = benchmark["4"]["chart"]
        query_meta = benchmark["4"]["query_meta"]
        result = chart_check(
            chart_info,
            ground_truth,
            query_meta[0]["stacked_bar"] if "stacked_bar" in query_meta[0] else None,
        )
        assert result[0] is True


def test_chart_check_scatter_400():
    for i in range(2):
        with open(folder / f"assets/scatter_400_{i}.svg", "r") as f:
            svg_string = f.read()
            chart_info, msg = deconstruct(svg_string)
            ground_truth = benchmark["400"]["chart"]
            query_meta = benchmark["400"]["query_meta"]
            result = chart_check(
                chart_info,
                ground_truth,
                (
                    query_meta[i]["stacked_bar"]
                    if "stacked_bar" in query_meta[i]
                    else None
                ),
            )
            assert result[0] is True


def test_chart_check_bar_1129():
    for i in range(3):
        with open(folder / f"assets/bar_1129_{i}.svg", "r") as f:
            svg_string = f.read()
            chart_info, msg = deconstruct(svg_string)
            ground_truth = benchmark["1129"]["chart"]
            query_meta = benchmark["1129"]["query_meta"]
            result = chart_check(
                chart_info,
                ground_truth,
                (
                    query_meta[i]["stacked_bar"]
                    if "stacked_bar" in query_meta[i]
                    else None
                ),
            )
            assert result[0] is True


def test_chart_check_bar_2750():
    for i in range(2):
        with open(folder / f"assets/stacked_bar_2750_{i}.svg", "r") as f:
            svg_string = f.read()
            chart_info, msg = deconstruct(svg_string)
            ground_truth = benchmark["2750"]["chart"]
            query_meta = benchmark["2750"]["query_meta"]
            result = chart_check(
                chart_info,
                ground_truth,
                (
                    query_meta[i]["stacked_bar"]
                    if "stacked_bar" in query_meta[i]
                    else None
                ),
            )
            assert result[0] is True


def test_chart_check_line_3240():
    with open(folder / f"assets/line_3240_0.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        ground_truth = benchmark["3240"]["chart"]
        query_meta = benchmark["3240"]["query_meta"]
        result = chart_check(
            chart_info,
            ground_truth,
            query_meta[0]["stacked_bar"] if "stacked_bar" in query_meta[0] else None,
        )
        assert result[0] is True


def test_chart_check_line_2781():
    with open(folder / f"assets/grouping_line_2781_0.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        ground_truth = benchmark["2781"]["chart"]
        query_meta = benchmark["2781"]["query_meta"]
        result = chart_check(
            chart_info,
            ground_truth,
            query_meta[0]["stacked_bar"] if "stacked_bar" in query_meta[0] else None,
        )
        assert result[0] is True


def test_chart_check_bar_1071():
    # horizontal bar
    with open(folder / f"assets/bar_1071_0.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        ground_truth = benchmark["1071"]["chart"]
        query_meta = benchmark["1071"]["query_meta"]
        result = chart_check(
            chart_info,
            ground_truth,
            query_meta[0]["stacked_bar"] if "stacked_bar" in query_meta[0] else None,
        )
        assert result[0] is True


def test_stacked():
    result = chart_check({"chart": "grouping bar"}, "Stacked Bar", True)
    assert result[0] is False

    result = chart_check({"chart": "grouping bar"}, "Stacked Bar", False)
    assert result[0] is True



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import json
from pathlib import Path


class Dataset:
    def __init__(
        self,
        folder: Path,
        table_type: str = "all",
        with_irrelevant_tables: bool = False,
    ):
        self.folder = folder
        dict_name = "visEval"
        if table_type in ["single", "multiple"]:
            dict_name += "_" + table_type
        dict_name += ".json"
        with open(folder / dict_name) as f:
            self.dict = json.load(f)

        with open(folder / "databases/db_tables.json") as f:
            self.db_tables = json.load(f)

        def benchmark():
            for key in list(self.dict.keys()):
                self.dict[key]["id"] = key
                self.dict[key]["tables"] = self.__get_tables(
                    key, with_irrelevant_tables
                )
                yield self.dict[key]

        self.benchmark = benchmark()

    def __get_tables(self, id: str, with_irrelevant_tables: bool = False):
        spec = self.dict[id]
        db_id = spec["db_id"]
        # table name
        all_table_names = self.db_tables[db_id]
        table_names = [
            x
            for x in all_table_names
            if x.lower() in spec["vis_query"]["VQL"].lower().split()
        ]

        if with_irrelevant_tables:
            irrelevant_tables = spec["irrelevant_tables"]
            table_names.extend(irrelevant_tables)

        tables = list(
            map(
                lambda table_name: f"{self.folder}/databases/{db_id}/{table_name}.csv",
                table_names,
            )
        )

        return tables



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

from pathlib import Path

import pytest

from viseval.check import deconstruct

folder = Path(__file__).resolve().parent


# error case
def test_deconstruct_empty():
    with open(folder / "assets/empty.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert len(chart_info["data"]) == 0


def test_deconstruct_empty1():
    with open(folder / "assets/empty_1.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert len(chart_info["data"]) == 0


def test_deconstruct_empty2():
    with open(folder / "assets/empty_2.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert len(chart_info["data"]) == 0


def test_deconstruct_empty3():
    with open(folder / "assets/empty_3.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info is None


def test_deconstruct_empty4():
    with open(folder / "assets/empty_4.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        print(chart_info)
        assert len(chart_info["data"]) == 0


def test_deconstruct_dual_axis():
    with open(folder / "assets/dual_1499_0.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info is None


def test_deconstruct_bar_error():
    with open(folder / "assets/bar_error_3227_0.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "bar"
        assert chart_info["chart"] == "vertical bar(error)"


def test_deconstruct_scatter():
    with open(folder / "assets/scatter.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "circle"
        assert chart_info["chart"] == "scatter"
        assert len(chart_info["encoding"].keys()) == 2
        assert (
            chart_info["title"]
            == "Correlation between IMDB Rating and Rotten Tomatoes Rating in Action Movies"
        )
        assert chart_info["encoding"]["x"]["title"] == "IMDB Rating"
        assert chart_info["encoding"]["y"]["title"] == "Rotten Tomatoes Rating"
        assert len(chart_info["data"]) == 309


def test_deconstruct_scatter1():
    with open(folder / "assets/scatter_1.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "circle"
        assert chart_info["chart"] == "scatter"
        assert len(chart_info["encoding"].keys()) == 2
        assert len(chart_info["data"]) == 5


def test_deconstruct_scatter2():
    with open(folder / "assets/scatter_2.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "circle"
        assert chart_info["chart"] == "scatter"
        assert len(chart_info["encoding"].keys()) == 2
        assert len(chart_info["data"]) == 100


def test_deconstruct_scatter4():
    with open(folder / "assets/scatter_4.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "circle"
        assert chart_info["chart"] == "scatter"
        assert len(chart_info["encoding"].keys()) == 2
        assert len(chart_info["data"]) == 3


def test_deconstruct_scatter6():
    with open(folder / "assets/scatter_6.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "circle"
        assert chart_info["chart"] == "scatter"
        assert len(chart_info["encoding"].keys()) == 2
        assert len(chart_info["data"]) == 69


def test_deconstruct_scatter_lida():
    with open(folder / "assets/scatter_62.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "circle"
        assert chart_info["chart"] == "scatter"
        assert len(chart_info["encoding"].keys()) == 2
        assert len(chart_info["data"]) == 9


def test_deconstruct_scatter_lida2():
    with open(folder / "assets/scatter_292.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "circle"
        assert chart_info["chart"] == "scatter"
        assert len(chart_info["encoding"].keys()) == 2
        assert len(chart_info["data"]) == 3


def test_deconstruct_scatter_lida3():
    with open(folder / "assets/scatter_674.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        print(chart_info)
        assert chart_info["mark"] == "circle"
        assert chart_info["chart"] == "scatter"
        assert len(chart_info["encoding"].keys()) == 2
        assert len(chart_info["data"]) == 76


# black legend
def test_deconstruct_grouping_scatter7():
    with open(folder / "assets/grouping_scatter_3272_0.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "circle"
        assert chart_info["chart"] == "grouping scatter"
        assert len(chart_info["encoding"].keys()) == 3
        assert len(chart_info["data"]) == 6


def test_deconstruct_grouping_scatter():
    with open(folder / "assets/grouping_scatter.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "circle"
        assert chart_info["chart"] == "grouping scatter"
        assert len(chart_info["encoding"].keys()) == 3
        assert len(chart_info["data"]) == 5


def test_deconstruct_line():
    with open(folder / "assets/line.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "line"
        assert chart_info["chart"] == "line"
        assert len(chart_info["encoding"].keys()) == 2
        assert chart_info["title"] == "Maximum Temperature by Month"
        assert chart_info["encoding"]["x"]["title"] == "Month"
        assert chart_info["encoding"]["y"]["title"] == "Temperature (°C)"
        assert len(chart_info["data"]) == 12


def test_deconstruct_line1():
    with open(folder / "assets/line_1.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "line"
        assert chart_info["chart"] == "line"
        assert len(chart_info["encoding"].keys()) == 2
        assert len(chart_info["data"]) == 20


def test_deconstruct_line2():
    with open(folder / "assets/line_2.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "line"
        assert chart_info["chart"] == "line"
        assert len(chart_info["encoding"].keys()) == 2
        assert len(chart_info["data"]) == 107


def test_deconstruct_line3():
    with open(folder / "assets/line_3.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "line"
        assert chart_info["chart"] == "line"
        assert len(chart_info["encoding"].keys()) == 2
        assert len(chart_info["data"]) == 10


def test_deconstruct_line4():
    with open(folder / "assets/line_4.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "line"
        assert chart_info["chart"] == "line"
        assert len(chart_info["encoding"].keys()) == 2
        assert len(chart_info["data"]) == 107


# error case
def test_deconstruct_line5():
    with open(folder / "assets/line_5.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "line"
        assert chart_info["chart"] == "line"
        assert len(chart_info["encoding"].keys()) == 2
        assert len(chart_info["data"]) == 0


# error case
def test_deconstruct_line11():
    with open(folder / "assets/line_11.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert "mark" not in chart_info


# tick line
def test_deconstruct_line12():
    with open(folder / "assets/line_12.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "line"
        assert chart_info["chart"] == "line"
        assert len(chart_info["encoding"].keys()) == 2
        assert len(chart_info["data"]) == 13


def test_deconstruct_line13():
    with open(folder / "assets/line_13.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "line"
        assert chart_info["chart"] == "line"
        assert len(chart_info["encoding"].keys()) == 2
        assert len(chart_info["data"]) == 15


def test_deconstruct_line14():
    with open(folder / "assets/line_1746_1.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "circle"
        assert chart_info["chart"] == "scatter"
        assert len(chart_info["encoding"].keys()) == 2
        assert len(chart_info["data"]) == 1


def test_deconstruct_grouping_line():
    with open(folder / "assets/grouping_line.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "line"
        assert chart_info["chart"] == "grouping line"
        assert len(chart_info["encoding"].keys()) == 3
        assert len(chart_info["data"]) == 9


def test_deconstruct_grouping_line1():
    with open(folder / "assets/grouping_line_1.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "line"
        assert chart_info["chart"] == "grouping line"
        assert len(chart_info["encoding"].keys()) == 3
        assert len(chart_info["data"]) == 20


def test_deconstruct_grouping_line4():
    with open(folder / "assets/grouping_line_4.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert len(chart_info["encoding"].keys()) == 2


# error case
def test_deconstruct_grouping_line2():
    with open(folder / "assets/grouping_line_2.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "line"
        assert chart_info["chart"] == "grouping line"
        assert len(chart_info["encoding"].keys()) == 3
        assert len(chart_info["data"]) == 0


def test_deconstruct_grouping_line3():
    with open(folder / "assets/grouping_line_3.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "line"
        assert chart_info["chart"] == "grouping line"
        assert len(chart_info["encoding"].keys()) == 3
        assert len(chart_info["data"]) == 0


def test_deconstruct_bar():
    with open(folder / "assets/bar.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "bar"
        assert chart_info["chart"] == "vertical bar"
        assert len(chart_info["encoding"].keys()) == 2
        assert chart_info["title"] == "Number of Students Advised by Faculty Rank"
        assert chart_info["encoding"]["x"]["title"] == "Faculty Rank"
        assert chart_info["encoding"]["y"]["title"] == "Number of Students Advised"
        assert len(chart_info["data"]) == 3


def test_deconstruct_bar1():
    with open(folder / "assets/bar_1.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "bar"
        assert chart_info["chart"] == "horizontal bar"
        assert len(chart_info["encoding"].keys()) == 2
        assert len(chart_info["data"]) == 7


def test_deconstruct_bar6():
    with open(folder / "assets/bar_6.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "bar"
        assert len(chart_info["encoding"].keys()) == 2
        assert len(chart_info["data"]) == 49


# 1 bar
def test_deconstruct_bar7():
    with open(folder / "assets/bar_2733_0.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "bar"
        assert chart_info["chart"] == "horizontal bar"
        assert len(chart_info["encoding"].keys()) == 2
        assert len(chart_info["data"]) == 1


# bar legend error
def test_deconstruct_bar8():
    with open(folder / "assets/stacked_bar_2815.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "bar"
        assert len(chart_info["encoding"].keys()) == 2


def test_deconstruct_bar9():
    with open(folder / "assets/bar_186.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "bar"
        assert len(chart_info["encoding"].keys()) == 2
        assert len(chart_info["data"]) == 14


def test_deconstruct_bar10():
    with open(folder / "assets/bar_3269.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "bar"
        assert len(chart_info["encoding"].keys()) == 2
        print(chart_info)
        assert len(chart_info["data"]) == 6


# lida bar
def test_deconstruct_bar_lida1():
    with open(folder / "assets/bar_9.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "bar"
        assert chart_info["chart"] == "vertical bar"
        assert len(chart_info["encoding"].keys()) == 2
        assert len(chart_info["data"]) == 4


# lida bar
def test_deconstruct_bar_lida2():
    with open(folder / "assets/bar_205.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "bar"
        assert chart_info["chart"] == "vertical bar"
        assert len(chart_info["encoding"].keys()) == 2
        assert len(chart_info["data"]) == 2


def test_deconstruct_stacked_bar():
    with open(folder / "assets/stacked_bar.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "bar"
        assert chart_info["chart"] == "vertical stacked bar"
        assert len(chart_info["encoding"].keys()) == 3
        assert chart_info["title"] == "Faculty by Rank/Sex"
        assert chart_info["encoding"]["x"]["title"] == "Rank"
        assert chart_info["encoding"]["y"]["title"] == "Number of faculty"
        assert len(chart_info["data"]) == 7


def test_deconstruct_stacked_bar1():
    with open(folder / "assets/stacked_bar_1.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "bar"
        assert chart_info["chart"] == "vertical stacked bar"
        assert len(chart_info["encoding"].keys()) == 3
        assert len(chart_info["data"]) == 9


def test_deconstruct_grouping_bar():
    with open(folder / "assets/bar_189.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "bar"
        assert chart_info["chart"] == "vertical grouping bar"
        assert len(chart_info["encoding"].keys()) == 3
        assert len(chart_info["data"]) == 4


def test_deconstruct_pie():
    with open(folder / "assets/pie.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "arc"
        assert chart_info["chart"] == "pie"
        assert len(chart_info["encoding"].keys()) == 2
        assert chart_info["title"] == "Maximum Price of Each Film"
        assert len(chart_info["data"]) == 5


# 180 degree
def test_deconstruct_pie1():
    with open(folder / "assets/pie_1.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "arc"
        assert chart_info["chart"] == "pie"
        assert len(chart_info["encoding"].keys()) == 2
        assert len(chart_info["data"]) == 3
        flag = False
        for datum in chart_info["data"]:
            # = 50 percent
            if datum["field_theta"] == 50:
                flag = True
        assert flag


# shadow and > 180 degree
def test_deconstruct_pie2():
    with open(folder / "assets/pie_2.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "arc"
        assert chart_info["chart"] == "pie"
        assert len(chart_info["encoding"].keys()) == 2
        assert len(chart_info["data"]) == 3
        flag = False
        for datum in chart_info["data"]:
            # > 50 percent
            if datum["field_theta"] > 50:
                flag = True
        assert flag


# shadow and > 180 degree
def test_deconstruct_pie3():
    with open(folder / "assets/pie_3.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "arc"
        assert chart_info["chart"] == "pie"
        assert len(chart_info["encoding"].keys()) == 2
        assert len(chart_info["data"]) == 2
        flag = False
        for datum in chart_info["data"]:
            # > 50 percent
            if datum["field_theta"] > 50:
                flag = True
        assert flag


# 360 degree
def test_deconstruct_pie3():
    with open(folder / "assets/pie_4.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        assert chart_info["mark"] == "arc"
        assert chart_info["chart"] == "pie"
        assert len(chart_info["encoding"].keys()) == 2
        assert len(chart_info["data"]) == 1
        flag = False
        for datum in chart_info["data"]:
            # > 50 percent
            if datum["field_theta"] == 100:
                flag = True
        assert flag



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.


def chart_check(
    chart_info: dict, chart_type_ground_truth: str, stacked_bar: bool = False
):
    # chart_type_ground_truth: pie, bar, line, scatter, stacked bar, grouping line, grouping scatter
    if "chart" not in chart_info:
        return False, "Cannot recognize the chart type."

    chart_type = chart_info["chart"]

    if chart_type_ground_truth.lower() in chart_type.lower() or (
        chart_type_ground_truth == "Stacked Bar"
        and not stacked_bar
        and "grouping bar" in chart_type.lower()
    ):
        return True, "Chart type is consistent with ground truth."

    return False, "Chart type is not consistent with ground truth."



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

from abc import ABC, abstractmethod
from typing import Optional, Tuple, Union

from attr import dataclass
from langchain.chat_models.base import BaseChatModel
from llmx import TextGenerator


@dataclass
class ChartExecutionResult:
    """Response from a visualization execution"""

    # True if successful, False otherwise
    status: bool
    # Generate svg string if status is True
    svg_string: Optional[str] = None
    # Error message if status is False
    error_msg: Optional[str] = None


class Agent(ABC):
    @abstractmethod
    def __init__(
        self, llm: Union[BaseChatModel, TextGenerator], config: dict = None
    ) -> None:
        pass

    @abstractmethod
    def generate(
        self, nl_query: str, tables: list[str], config: dict
    ) -> Tuple[str, dict]:
        """Generate code for the given natural language query.

        Args:
            nl_query (str): Natural language query.
            tables (list[str]): List of table file paths.
            config (dict): Generation configuration.

        Returns:
            Tuple[str, dict]: Generated code and execution context.
        """
        pass

    @abstractmethod
    def execute(
        self, code: str, context: dict, log_name: str = None
    ) -> ChartExecutionResult:
        """Execute the given code with context and return the result

        Args:
            code (str): Code to execute,.
            context (dict): Context for the code execution. Different agents require different contexts.
            log_name (str, optional): SVG file name. Defaults to None.

        Returns:
            ChartExecutionResult: _description_
        """
        pass



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import json
from pathlib import Path

import pytest

from viseval.check import data_check, deconstruct

folder = Path(__file__).resolve().parent

with open(folder / "assets/samples.json") as f:
    benchmark = json.load(f)


def test_data_check_pie_4():
    with open(folder / "assets/pie_4_0.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        ground_truth = benchmark["4"]["vis_obj"]
        query_meta = benchmark["4"]["query_meta"]
        result = data_check(
            chart_info, ground_truth, query_meta[0]["channel_specified"]
        )
        assert result[0] is True


def test_data_check_scatter_400():
    for i in range(2):
        with open(folder / f"assets/scatter_400_{i}.svg", "r") as f:
            svg_string = f.read()
            chart_info, msg = deconstruct(svg_string)
            ground_truth = benchmark["400"]["vis_obj"]
            query_meta = benchmark["400"]["query_meta"]
            result = data_check(
                chart_info, ground_truth, query_meta[i]["channel_specified"]
            )
            assert result[0] is True


def test_data_check_bar_1129():
    for i in range(3):
        with open(folder / f"assets/bar_1129_{i}.svg", "r") as f:
            svg_string = f.read()
            chart_info, msg = deconstruct(svg_string)
            ground_truth = benchmark["1129"]["vis_obj"]
            query_meta = benchmark["1129"]["query_meta"]
            result = data_check(
                chart_info, ground_truth, query_meta[i]["channel_specified"]
            )
            assert result[0] is True


def test_data_check_bar_2750():
    for i in range(2):
        with open(folder / f"assets/stacked_bar_2750_{i}.svg", "r") as f:
            svg_string = f.read()
            chart_info, msg = deconstruct(svg_string)
            ground_truth = benchmark["2750"]["vis_obj"]
            query_meta = benchmark["2750"]["query_meta"]
            result = data_check(
                chart_info, ground_truth, query_meta[i]["channel_specified"]
            )
            assert result[0] is True


def test_data_check_line_3240():
    with open(folder / "assets/line_3240_0.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        ground_truth = benchmark["3240"]["vis_obj"]
        query_meta = benchmark["3240"]["query_meta"]
        result = data_check(
            chart_info, ground_truth, query_meta[0]["channel_specified"]
        )
        assert result[0] is True


def test_data_check_line_2781():
    with open(folder / "assets/grouping_line_2781_0.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        ground_truth = benchmark["2781"]["vis_obj"]
        query_meta = benchmark["2781"]["query_meta"]
        result = data_check(
            chart_info, ground_truth, query_meta[0]["channel_specified"]
        )
        assert result[0] is False


def test_data_check_line_773():
    # yyyy-mm-dd
    with open(folder / "assets/line_773_0.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        ground_truth = benchmark["773@x_name@DESC"]["vis_obj"]
        query_meta = benchmark["773@x_name@DESC"]["query_meta"]
        result = data_check(
            chart_info, ground_truth, query_meta[0]["channel_specified"]
        )
        assert result[0] is True


def test_data_check_bar_68():
    # monday - mon
    with open(folder / "assets/bar_68_0.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        ground_truth = benchmark["68"]["vis_obj"]
        query_meta = benchmark["68"]["query_meta"]
        result = data_check(
            chart_info, ground_truth, query_meta[0]["channel_specified"]
        )
        assert result[0] is True


def test_data_check_bar_1071():
    # horizontal bar
    with open(folder / "assets/bar_1071_0.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        ground_truth = benchmark["1071"]["vis_obj"]
        query_meta = benchmark["1071"]["query_meta"]
        result = data_check(
            chart_info, ground_truth, query_meta[0]["channel_specified"]
        )
        assert result[0] is True

        result = data_check(chart_info, ground_truth, ["x", "y"])
        assert result[0] is False


def test_data_check_bar_3137():
    # error cases
    with open(folder / "assets/bar_3137_1.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        ground_truth = benchmark["3137@y_name@DESC"]["vis_obj"]
        query_meta = benchmark["3137@y_name@DESC"]["query_meta"]
        result = data_check(
            chart_info, ground_truth, query_meta[0]["channel_specified"]
        )
        assert result[0] is False



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

from .chart_check import chart_check
from .data_check import data_check
from .deconstruct import deconstruct
from .layout_check import layout_check
from .order_check import order_check
from .readability_check import readability_check
from .scale_and_ticks_check import scale_and_ticks_check



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

from __future__ import annotations

import json
import sys
import warnings

from langchain.chat_models.base import BaseChatModel
from langchain.schema import HumanMessage, SystemMessage

INSTRUCTION = """You will be provided with a visualization and its specifications. Consider the following aspect:

    - If the scale selected for the visualization is appropriate for accurate interpretation of values, avoid using unconventional scales, such as an inverted y-axis scale.
    - When axes are present, ensure that the selected ticks are appropriate for clarity, avoiding unconventional choices, such as representing counts of individual entities with floating-point ticks.


Report your findings, focusing solely on scale and tick appropriateness without considering the order.
```
    {
        "Appropriate": true/false,
        "Rationale": "reason ..."
    }
```
"""


def scale_and_ticks_check(context: dict, query: str, vision_model: BaseChatModel):
    base64 = context["base64"]
    encoding = context["encoding"]
    chart = context["chart"]
    if chart == "pie":
        ticks_desc = ""
    else:
        x_ticks = encoding["x"]["scale"]["ticks"]
        y_ticks = encoding["y"]["scale"]["ticks"]
        ticks_desc = f"Ticks extracted from the visualization:\n- x axis ticks: {','.join(x_ticks)}\n- y axis ticks: {','.join(y_ticks)}\n\n"
    try:
        messages = [
            SystemMessage(content=INSTRUCTION),
        ]
        messages.append(
            HumanMessage(
                content=[
                    {
                        "type": "text",
                        "text": f"Visualization specification:{query}\n\n{ticks_desc}Visualization image that has been verified for DATA and ORDER accuracy:",
                    },
                    {
                        "type": "image_url",
                        "image_url": base64,
                    },
                ]
            )
        )

        response = vision_model.invoke(messages)

        json_string = (
            response.content.replace("```json\n", "").replace("```", "").strip()
        )
        try:
            result = json.loads(json_string)
        except Exception:
            result = eval(json_string)
        return result["Appropriate"], result["Rationale"]
    except Exception:
        warnings.warn(str(sys.exc_info()))
    return None, "Exception occurred."



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import csv
from pathlib import Path

import pytest

from viseval.check import layout_check

folder = Path(__file__).resolve().parent

webdriver_path = "/usr/bin/chromedriver"


def test_layout_check():
    with open(folder / "assets/readability/145@y_name@DESC.svg", "r") as svg_file:
        svg_string = svg_file.read()
    assert layout_check({"svg_string": svg_string}, webdriver_path) == (
        False,
        "Overflow detected.",
    )


def test_layout_check_2():
    with open(folder / "assets/readability/2350.svg", "r") as svg_file:
        svg_string = svg_file.read()
    assert layout_check({"svg_string": svg_string}, webdriver_path) == (
        True,
        "No overflow or overlap detected.",
    )


def test_layout_check_3():
    with open(folder / "assets/readability/2652.svg", "r") as svg_file:
        svg_string = svg_file.read()
    assert layout_check({"svg_string": svg_string}, webdriver_path) == (
        False,
        "Overflow detected.",
    )


def test_layout_check_batch():
    with open(folder / "assets/readability/readability_human_rating.csv") as f:
        reader = csv.reader(f)
        next(reader)
        for row in reader:
            svg_id = row[0]

        with open(folder / f"assets/readability/{svg_id}.svg", "r") as svg_file:
            svg_string = svg_file.read()
            print(svg_id)
        assert not layout_check({"svg_string": svg_string}, webdriver_path)[0] == (
            False if row[5] == "FALSE" else True
        )



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import json
from pathlib import Path

import pytest

from viseval.check import data_check, deconstruct, order_check

folder = Path(__file__).resolve().parent

with open(folder / "assets/samples.json") as f:
    benchmark = json.load(f)


def test_order_check_line_773():
    # x ascending
    with open(folder / "assets/line_773_0.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        # yyyy-mm-dd
        ground_truth = benchmark["773@x_name@DESC"]["vis_obj"]
        query_meta = benchmark["773@x_name@DESC"]["query_meta"]
        answer, rationale = data_check(
            chart_info, ground_truth, query_meta[0]["channel_specified"]
        )

        answer, rationale = order_check(
            chart_info, ground_truth, query_meta[0]["sort_by"]
        )
        assert answer is True

        ground_truth["sort"]["order"] = "descending"
        answer, rationale = order_check(
            chart_info, ground_truth, query_meta[0]["sort_by"]
        )
        assert answer is False


def test_order_check_bar_680():
    # y descending
    with open(folder / "assets/stacked_bar_680_0.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        ground_truth = benchmark["680@y_name@DESC"]["vis_obj"]
        query_meta = benchmark["680@y_name@DESC"]["query_meta"]
        answer, rationale = data_check(
            chart_info, ground_truth, query_meta[0]["channel_specified"]
        )

        answer, rationale = order_check(
            chart_info, ground_truth, query_meta[0]["sort_by"]
        )
        assert answer is True

        ground_truth["sort"]["order"] = "ascending"
        answer, rationale = order_check(
            chart_info, ground_truth, query_meta[0]["sort_by"]
        )
        assert answer is False


def test_order_check_bar_2815():
    # x descending
    # swap channel x-z
    with open(folder / "assets/stacked_bar_2815_0.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        ground_truth = benchmark["2815@x_name@DESC"]["vis_obj"]
        query_meta = benchmark["2815@x_name@DESC"]["query_meta"]
        answer, rationale = data_check(
            chart_info, ground_truth, query_meta[0]["channel_specified"]
        )

        answer, rationale = order_check(
            chart_info, ground_truth, query_meta[0]["sort_by"]
        )
        assert answer is True

        ground_truth["sort"]["order"] = "ascending"
        answer, rationale = order_check(
            chart_info, ground_truth, query_meta[0]["sort_by"]
        )
        assert answer is False


# horizontal bar
def test_order_check_bar_2060():
    # y descending
    # swap channel x-y
    with open(folder / "assets/bar_2060_0.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        ground_truth = benchmark["2060@y_name@DESC"]["vis_obj"]
        query_meta = benchmark["2060@y_name@DESC"]["query_meta"]
        answer, rationale = data_check(chart_info, ground_truth, [])

        answer, rationale = order_check(
            chart_info, ground_truth, query_meta[0]["sort_by"]
        )
        assert answer is False

        # x quantitative descending
        answer, rationale = order_check(chart_info, ground_truth, "attribute")
        assert answer is True

        ground_truth["sort"]["order"] = "ascending"
        answer, rationale = order_check(chart_info, ground_truth, "attribute")
        assert answer is False


# horizontal bar
def test_order_check_bar_1000():
    # y descending
    # swap channel x-y
    with open(folder / "assets/bar_1000_0.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        ground_truth = benchmark["1000@y_name@DESC"]["vis_obj"]
        query_meta = benchmark["1000@y_name@DESC"]["query_meta"]
        answer, rationale = data_check(chart_info, ground_truth, [])
        print(answer, rationale)
        answer, rationale = order_check(
            chart_info, ground_truth, query_meta[0]["sort_by"]
        )
        assert answer is False

        # x quantitative descending
        ground_truth["sort"]["channel"] = "x"
        answer, rationale = order_check(
            chart_info, ground_truth, query_meta[0]["sort_by"]
        )
        assert answer is True

        # y nominal descending
        ground_truth["sort"]["channel"] = "y"
        chart_info["encoding"]["y"]["scale"]["domain"] = [
            "sony",
            "jcrew",
            "gucci",
            "apple",
        ]
        answer, rationale = order_check(
            chart_info, ground_truth, query_meta[0]["sort_by"]
        )
        assert answer is True

        # custom order
        ground_truth["sort"]["order"] = ["sony", "jcrew", "gucci", "apple"]
        answer, rationale = order_check(
            chart_info, ground_truth, query_meta[0]["sort_by"]
        )
        assert answer is True


# invert y axis, zero value
def test_order_check_bar_550():
    with open(folder / "assets/bar_550_0.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        ground_truth = benchmark["550@y_name@DESC"]["vis_obj"]
        query_meta = benchmark["550@y_name@DESC"]["query_meta"]
        answer, rationale = data_check(
            chart_info, ground_truth, query_meta[0]["channel_specified"]
        )

        answer, rationale = order_check(
            chart_info, ground_truth, query_meta[0]["sort_by"]
        )
        assert answer is True


def test_order_check_bar_465():
    with open(folder / "assets/bar_465.svg", "r") as f:
        svg_string = f.read()
        chart_info, msg = deconstruct(svg_string)
        ground_truth = benchmark["465@y_name@ASC"]["vis_obj"]
        query_meta = benchmark["465@y_name@ASC"]["query_meta"]
        answer, rationale = data_check(
            chart_info, ground_truth, query_meta[0]["channel_specified"]
        )

        answer, rationale = order_check(
            chart_info, ground_truth, query_meta[0]["sort_by"]
        )
        assert answer is False



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import os
import uuid

overflowScript = """const isElementOutsideCanvas = (element) => {
    // Get the bounding box of the SVG element
    let svgRect = element.getBBox();

    // Get the viewBox attribute of the SVG element
    let svgElement = document.querySelector('svg');
    let viewBoxValue = svgElement.getAttribute('viewBox');
    let viewBoxArray = viewBoxValue.split(' ').map(Number);
    let minX = viewBoxArray[0];
    let minY = viewBoxArray[1];
    let width = viewBoxArray[2];
    let height = viewBoxArray[3];

    // Check if any part of the element is outside the canvas
    if (svgRect.x < minX || svgRect.y < minY || svgRect.x + svgRect.width > width+minX || svgRect.y + svgRect.height > height+minY) {
        return true;
    } else {
        return false;
    }
}

let svgElement = document.querySelector('#axes_1');
return isElementOutsideCanvas(svgElement);
"""

overlapScript = """const isTextOverlap = (parentElement) => {
    let index = 1;
    let flag = true;
    let textArray = [];
    while (flag) {
        let textElement = parentElement.querySelector('#text_' + index);
        if (textElement) {
            index++;
            let gElement = textElement.querySelector('g');
            let transform = gElement.getAttribute('transform');
            if (transform) {
                const rotateMatch = transform.match(/rotate\(([^)]+)\)/);
                let notRotate = true;
                if (rotateMatch) {
                    const rotateValue = parseFloat(rotateMatch[1]) % 90;
                    if (Math.abs(rotateValue) > 10 && Math.abs(rotateValue) < 80) {
                        notRotate = false;
                    }
                }

                if (notRotate) {
                    let bbox = textElement.getBBox();
                    textArray.push(bbox);
                }
            }
        }
        else {
            flag = false;
        }
    }
    let textOverlap = false;
    let textArrayLength = textArray.length;
    for (let i = 0; i < textArrayLength; i++) {
        for (let j = i + 1; j < textArrayLength; j++) {
            let text1Rect = textArray[i];
            let text2Rect = textArray[j];
            if (text1Rect.x < text2Rect.x + text2Rect.width &&
                text1Rect.x + text1Rect.width > text2Rect.x &&
                text1Rect.y < text2Rect.y + text2Rect.height &&
                text1Rect.y + text1Rect.height > text2Rect.y) {
                textOverlap = true;
                break;
            }
        }
    }
    return textOverlap;
}

let svgElement = document.querySelector('#axes_1');
return isTextOverlap(svgElement)
"""


def layout_check(context: dict, webdriver_path):
    svg_string = context["svg_string"]

    if webdriver_path is not None:
        from selenium import webdriver
        from selenium.webdriver.chrome.service import Service

        # Chrome headless mode
        options = webdriver.ChromeOptions()
        options.add_argument("--headless")
        options.add_argument("--disable-gpu")
        try:
            service = Service(webdriver_path)
            webdriver = webdriver.Chrome(service=service, options=options)

            current_directory = os.getcwd()
            file_path = f"{current_directory}/temp_{uuid.uuid1()}.svg"

            with open(file_path, "w") as svg_file:
                svg_file.write(svg_string)

            webdriver.get(f"file://{file_path}")
            overflow_result = not webdriver.execute_script(overflowScript)
            overlap_result = not webdriver.execute_script(overlapScript)
            webdriver.close()
            try:
                os.remove(file_path)
            except Exception as e:
                print(f"Remove file error: {e}")

            if overflow_result and overlap_result:
                msg = "No overflow or overlap detected."
            elif not overflow_result and not overlap_result:
                msg = "Overflow and overlap detected."
            elif not overflow_result:
                msg = "Overflow detected."
            else:
                msg = "Overlap detected."
            return overflow_result and overlap_result, msg
        except Exception as e:
            print(e)

    return None, "No webdriver path provided."



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

from __future__ import annotations

import json
import sys
import warnings

from langchain.chat_models.base import BaseChatModel
from langchain.schema import HumanMessage, SystemMessage

INSTRUCTION = """Your task is to evaluate the readability of the visualization on a scale of 1 to 5, where 1 indicates very difficult to read and 5 indicates very easy to read. You will be given a visualization requirement and the corresponding visualization created based on that requirement. Additionally, reviews from others regarding this visualization will be provided for your reference. Please think carefully and provide your reasoning and score.
```
    {
        "Rationale": "a brief reason",
        "Score": 1-5
    }
```


Examples:
- If the visualization is clear and information can be easily interpreted, you might return:
```
    {
        "Rationale": "The chart is well-organized, and the use of contrasting colors helps in distinguishing different data sets effectively. The labels are legible, and the key insights can be understood at a glance.",
        "Score": 5
    }
```
- Conversely, if the visualization is cluttered or confusing, you might return:
```
    {
        "Rationale": "While there is no overflow or overlap, the unconventional inverted y-axis and the use of decimal numbers for months on the x-axis deviate from the standard interpretation of bar charts, confusing readers and significantly affecting the chart's readability.",
        "Score": 1
    }
```
"""


def readability_check(context: dict, query: str, vision_model: BaseChatModel):
    base64 = context["base64"]

    reviews = ""
    if "reviews" in context and len(context["reviews"]) > 0:
        reviews = "Other Reviews:\n"
        reviews += "\n".join(
            [
                f"""- {review["aspect"]}: {review["content"]}"""
                for review in context["reviews"]
            ]
        )
        reviews += "\n\n"

    try:
        messages = [
            SystemMessage(content=INSTRUCTION),
        ]
        messages.append(
            HumanMessage(
                content=[
                    {
                        "type": "text",
                        "text": f"""Visualization Requirement: {query}\n\n{reviews}Visualization image:""",
                    },
                    {
                        "type": "image_url",
                        "image_url": base64,
                    },
                    {
                        "type": "text",
                        "text": """Please assess the readability, taking into account factors such as layout, scale and ticks, title and labels, colors, and ease of extracting information. Do not consider the correctness of the data and order in the visualizations, as they have already been verified.""",
                    },
                ]
            )
        )

        response = vision_model.invoke(messages)

        json_string = (
            response.content.replace("```json\n", "").replace("```", "").strip()
        )
        try:
            result = json.loads(json_string)
        except Exception:
            result = eval(json_string)

        return result["Score"], result["Rationale"]
    except Exception:
        warnings.warn(str(sys.exc_info()))
    return None, "Exception occurred."



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import copy
import json

from .time_utils import (
    compare_time_strings,
    convert_month_or_weekday_to_int,
    is_month_or_weekday,
    parse_number_to_time,
    parse_time_to_timestamp,
)

PRECISION = 0.0005


def is_numeric(s):
    try:
        float(s)
        return True
    except ValueError:
        return False


def convert_ground_truth_data(ground_truth):
    # ground truth data
    x_data = ground_truth["x_data"]
    y_data = ground_truth["y_data"]
    classify = ground_truth["classify"]

    data_ground_truth = []
    if len(classify) == 0:
        x_data = x_data[0]
        y_data = y_data[0]
        for index in range(len(x_data)):
            data_ground_truth.append(
                {
                    "field_x": x_data[index],
                    "field_y": y_data[index],
                }
            )
    else:
        # scatter
        if len(classify) == len(x_data) and len(classify) == len(y_data):
            for index in range(len(classify)):
                for index2 in range(len(x_data[index])):
                    data_ground_truth.append(
                        {
                            "field_x": x_data[index][index2],
                            "field_y": y_data[index][index2],
                            "field_classify": classify[index],
                        }
                    )
        # line / bar
        elif len(classify) == len(y_data):
            for index in range(len(classify)):
                for index2 in range(len(x_data[0])):
                    data_ground_truth.append(
                        {
                            "field_x": x_data[0][index2],
                            "field_y": y_data[index][index2],
                            "field_classify": classify[index],
                        }
                    )
    return data_ground_truth


# return answer and rationale for the data check
def compare_data(data_ground_truth, chart_info):
    # deep copy: avoid to change origin data
    data = copy.deepcopy(chart_info["data"])
    encoding = chart_info["encoding"]
    # line chart may have different length of data because line chart might omit some values
    if (len(data) != len(data_ground_truth) and chart_info["mark"] != "line") or (
        len(data) > len(data_ground_truth)
    ):
        return (
            False,
            f"visualization data length {len(data)} != ground truth length {len(data_ground_truth)}.",
        )

    if chart_info["mark"] == "arc":
        # field_fill -> field_x
        field_x = "field_fill"
        # field_theta -> field_y relative
        field_y = "field_theta"

        scale = None
        for datum_ground_truth in data_ground_truth:
            datum = [
                x
                for x in data
                if str(x[field_x]).strip() == str(datum_ground_truth["field_x"]).strip()
            ]
            if len(datum) == 1:
                datum = datum[0]
                if scale is None:
                    scale = datum[field_y] / datum_ground_truth["field_y"]
                else:
                    if (
                        abs(datum[field_y] / datum_ground_truth["field_y"] - scale)
                        > PRECISION
                    ):
                        return False, f"{json.dumps(datum_ground_truth)} not found\n"
            elif len(datum) == 0:
                return False, f"{json.dumps(datum_ground_truth)} not found."
            elif len(datum) > 1:
                return False, f"{json.dumps(datum_ground_truth)} found more than one."
    else:
        if len(encoding.keys()) < len(data_ground_truth[0].keys()):
            return False, "too few channels\n"
        elif len(encoding.keys()) > 3:
            return False, "too many channels\n"
        elif len(encoding.keys()) > len(data_ground_truth[0].keys()):
            # only keep x and y for comparison
            encoding = {key: encoding[key] for key in ["x", "y"]}

        for datum_ground_truth in data_ground_truth:
            datum = data
            for key in encoding:
                if key == "x" or key == "y":
                    field = f"field_{key}"
                else:
                    field = "field_classify"
                if (
                    encoding[key]["type"] == "quantitative"
                    or encoding[key]["type"] == "temporal"
                ):
                    # e.g., Monday -> 1
                    if is_month_or_weekday(datum_ground_truth[field]):
                        if encoding[key]["type"] == "temporal":
                            datum = [
                                x
                                for x in datum
                                if compare_time_strings(
                                    x["field_" + key].strip(),
                                    str(datum_ground_truth[field]).strip(),
                                )
                            ]
                            value_ground_truth = None
                        else:
                            value_ground_truth = convert_month_or_weekday_to_int(
                                datum_ground_truth[field]
                            )
                    else:
                        try:
                            value_ground_truth = float(datum_ground_truth[field])
                        except Exception:
                            # not a number
                            # try temporal
                            try:
                                value_ground_truth = parse_time_to_timestamp(
                                    datum_ground_truth[field]
                                )
                                if value_ground_truth is None:
                                    return (
                                        False,
                                        f"The data type of {key}({encoding[key]['type']}) is wrong.",
                                    )

                                if encoding[key]["type"] == "quantitative":
                                    for x in datum:
                                        x["field_" + key] = parse_number_to_time(
                                            x["field_" + key]
                                        )
                            except Exception:
                                return (
                                    False,
                                    f"The data type of {key}({encoding[key]['type']}) is wrong.",
                                )

                    field_vis = (
                        "field_" + key
                        if encoding[key]["type"] != "temporal"
                        else "field_" + key + "_origin"
                    )
                    if value_ground_truth is not None:
                        # avoid division by zero
                        datum = [
                            x
                            for x in datum
                            if abs((x[field_vis] - value_ground_truth)) <= PRECISION
                            or (
                                value_ground_truth != 0
                                and abs(
                                    (x[field_vis] - value_ground_truth)
                                    / value_ground_truth
                                )
                                <= PRECISION
                            )
                        ]
                elif encoding[key]["type"] == "nominal":
                    # exact match or time match
                    datum = [
                        x
                        for x in datum
                        if x["field_" + key].strip()
                        == str(datum_ground_truth[field]).strip()
                        or compare_time_strings(
                            x["field_" + key].strip(),
                            str(datum_ground_truth[field]).strip(),
                        )
                    ]

            if len(datum) == 0:
                if (
                    chart_info["mark"] == "line"
                    and (
                        encoding["x"]["type"] == "quantitative"
                        or encoding["x"]["type"] == "temporal"
                    )
                    and (encoding["y"]["type"] == "quantitative")
                ):
                    # use all data
                    datum = chart_info["data"]
                    if len(data_ground_truth[0].keys()) == 3:
                        datum = [
                            x
                            for x in datum
                            if x["field_stroke"].strip()
                            == datum_ground_truth["field_classify"].strip()
                        ]
                    # line chart might omit some values
                    min_larger_index = -1
                    max_smaller_index = -1
                    # convert
                    field_vis = (
                        "field_x"
                        if encoding["x"]["type"] != "temporal"
                        else "field_x_origin"
                    )
                    value_ground_truth = datum_ground_truth["field_x"]
                    try:
                        value_ground_truth = float(value_ground_truth)
                    except Exception:
                        value_ground_truth = parse_time_to_timestamp(value_ground_truth)

                    for index in range(len(datum)):
                        if datum[index][field_vis] > value_ground_truth:
                            if (
                                min_larger_index == -1
                                or datum[index][field_vis]
                                < datum[min_larger_index][field_vis]
                            ):
                                min_larger_index = index
                        elif datum[index][field_vis] < value_ground_truth:
                            if (
                                max_smaller_index == -1
                                or datum[index][field_vis]
                                > datum[max_smaller_index][field_vis]
                            ):
                                max_smaller_index = index
                    if min_larger_index != -1 and max_smaller_index != -1:
                        if (
                            abs(
                                (
                                    datum[min_larger_index]["field_y"]
                                    - datum[max_smaller_index]["field_y"]
                                )
                            )
                            <= PRECISION
                            or abs(
                                (
                                    datum[min_larger_index]["field_y"]
                                    - datum_ground_truth["field_y"]
                                )
                            )
                            <= PRECISION
                        ):
                            continue
                        else:
                            return (
                                False,
                                f"{json.dumps(datum_ground_truth)} not found\n",
                            )
                return False, f"{json.dumps(datum_ground_truth)} not found."
            else:
                data.remove(datum[0])

    return True, "The data on the charts is consistent with the ground truth."


def data_check(chart_info: dict, data: dict, channel_specified: list):
    if ("data" not in chart_info) or (len(chart_info["data"]) == 0):
        return False, "The data on the charts cannot be understood."

    data_ground_truth = convert_ground_truth_data(data)
    # filter zero data
    if chart_info["mark"] == "bar":
        data_ground_truth = list(
            filter(lambda x: x["field_x"] != 0 and x["field_y"] != 0, data_ground_truth)
        )
        chart_info["data"] = list(
            filter(
                lambda x: x["field_x"] != 0 and x["field_y"] != 0,
                chart_info["data"],
            )
        )
    candidates = [data_ground_truth]
    # ground truth channel -> chart channel
    channel_maps = []
    if len(data["classify"]) == 0:
        # 2 channels
        channel_maps.append({"x": "x", "y": "y"})
        if "x" not in channel_specified and "y" not in channel_specified:
            # swap x and y
            data_ground_truth_copy = copy.deepcopy(data_ground_truth)
            for datum in data_ground_truth_copy:
                datum["field_x"], datum["field_y"] = (
                    datum["field_y"],
                    datum["field_x"],
                )
            candidates.append(data_ground_truth_copy)
            channel_maps.append({"x": "y", "y": "x"})
    else:
        # 3 channels
        channel_maps.append({"x": "x", "y": "y", "classify": "classify"})
        channels = ["x", "y", "classify"]
        if len(channel_specified) <= 1:
            swap_channels = list(set(channels) - set(channel_specified))
            data_ground_truth_copy = copy.deepcopy(data_ground_truth)
            for datum in data_ground_truth_copy:
                (
                    datum[f"field_{swap_channels[0]}"],
                    datum[f"field_{swap_channels[1]}"],
                ) = (
                    datum[f"field_{swap_channels[1]}"],
                    datum[f"field_{swap_channels[0]}"],
                )
            candidates.append(data_ground_truth_copy)
            channel_map = {"x": "x", "y": "y", "classify": "classify"}
            channel_map[swap_channels[0]], channel_map[swap_channels[1]] = (
                channel_map[swap_channels[1]],
                channel_map[swap_channels[0]],
            )
            channel_maps.append(channel_map)
        if len(channel_specified) == 0:
            # swap x and z
            data_ground_truth_copy = copy.deepcopy(candidates[0])
            for datum in data_ground_truth_copy:
                datum["field_x"], datum["field_classify"] = (
                    datum["field_classify"],
                    datum["field_x"],
                )
            candidates.append(data_ground_truth_copy)
            channel_maps.append({"x": "classify", "y": "y", "classify": "x"})
            # swap x and y, x and z
            data_ground_truth_copy = copy.deepcopy(candidates[1])
            for datum in data_ground_truth_copy:
                datum["field_y"], datum["field_classify"] = (
                    datum["field_classify"],
                    datum["field_y"],
                )
            candidates.append(data_ground_truth_copy)
            channel_maps.append({"x": "y", "y": "classify", "classify": "x"})
    cache = None
    for i in range(len(candidates)):
        candidate = candidates[i]
        channel_map = channel_maps[i]
        answer, rationale = compare_data(candidate, chart_info)
        if answer:
            chart_info["channel_map"] = channel_map
            return answer, rationale
        if i == 0:
            cache = [answer, rationale]

    return cache[0], cache[1]



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

# deconstruct svg to spec
import html
import math
import re
from collections import Counter
from xml.dom import minidom

import numpy as np

from .time_utils import is_datetime, parse_time_to_timestamp, parse_timestamp_to_time

keys_of_interest = [
    "id",
    "style",
    "transform",
    "translate",
    "rotate",
    "skewX",
    "skewY",
    "x",
    "dx",
    "dy",
    "y",
    "cx",
    "cy",
    "r",
    "x1",
    "x2",
    "y1",
    "y2",
    "width",
    "height",
    "fill",
    "color",
    "opacity",
    "fill-opacity",
    "stroke",
    "stroke-width",
    "stroke-opacity",
    "font-size",
    "font-weight",
    "text-anchor",
    "aria-label",
    "display",
    "matrix",
]
marks = ["bar", "arc", "circle", "line", "text"]
channels = ["fill", "width", "height", "r", "stroke", "stroke-width"]
PRECISION = 0.005


def extract_tag(node):
    return node.tagName


def get_attribute_names(node):
    return node.attributes.keys()


def get_attribute_value(node, attr):
    return node.getAttribute(attr)


def get_text_width(node, scale):
    width = 0
    for child in node.childNodes:
        if child.nodeType == child.ELEMENT_NODE and extract_tag(child) == "use":
            x = get_attribute_value(child, "x")
            if x != "":
                width = max(float(x), width)
    return width * scale


def process_text_matplotlib(node, spec, acc_trans):
    # Pass Text node
    is_first_line = True
    for child in node.childNodes:
        if child.nodeType != child.TEXT_NODE:
            # the fist node is comment node
            if child.nodeType == child.COMMENT_NODE:
                if "content" in spec:
                    if child.data.strip():
                        spec["content"] += "\n" + html.unescape(
                            child.data.strip()
                        ).replace("–", "-").replace("’", "'")
                else:
                    spec["content"] = (
                        html.unescape(child.data.strip())
                        .replace("–", "-")
                        .replace("’", "'")
                    )
                    spec["tag"] = "text"
            if child.nodeType == child.ELEMENT_NODE and is_first_line:
                if extract_tag(child) == "g":
                    # process transform
                    process_transform(child, spec, acc_trans)
                    if "translate" not in spec:
                        if "text" in get_attribute_value(child, "id"):
                            child_spec = {}
                            process_text_matplotlib(child, child_spec, acc_trans)
                            spec.update(child_spec)
                        continue
                    else:
                        if "x" not in spec:
                            spec["x"] = 0
                        else:
                            spec["x"] = float(spec["x"])
                        if "y" not in spec:
                            spec["y"] = 0
                        else:
                            spec["y"] = float(spec["y"])

                        spec["x"] += spec["translate"][0]
                        spec["y"] += spec["translate"][1]
                        del spec["translate"]
                        is_first_line = False
                        spec["width"] = get_text_width(child, spec["scale"][0])


def process_path(node, spec):
    allowed_operator = [
        "M",
        "m",
        "L",
        "l",
        "V",
        "v",
        "H",
        "h",
        "Z",
        "z",
        "A",
        "a",
        "C",
        "c",
    ]
    draw_seq = get_attribute_value(node, "d")
    point_seq = []
    char_seq = ""
    notCircle = (
        "L" in draw_seq
        or "l" in draw_seq
        or "V" in draw_seq
        or "v" in draw_seq
        or "H" in draw_seq
        or "h" in draw_seq
    )

    for i in range(len(draw_seq)):
        char = draw_seq[i]
        if char in allowed_operator:
            char_seq += char
            # close path
            if char == "Z" or char == "z":
                if len(point_seq) == 0:
                    break
                if (
                    point_seq[0][0] != point_seq[-1][0]
                    or point_seq[0][1] != point_seq[-1][1]
                ):
                    point_seq.append(point_seq[0])
            else:
                j = i + 1
                # drawSeq[j] is not letter
                while re.match(r"[^a-zA-Z]", draw_seq[j]):
                    j += 1
                    if j >= len(draw_seq):
                        break
                shifts = draw_seq[i + 1 : j].strip()
                if "," in shifts:
                    separator = shifts.index(",")
                else:
                    separator = -1
                if separator != -1:
                    af = list(filter(lambda d: d != "", shifts.split(",")))
                else:
                    af = list(filter(lambda d: d != "", shifts.split(" ")))
                if len(point_seq) > 0:
                    previous_x = point_seq[-1][0]
                    previous_y = point_seq[-1][1]

                if char == "A" and not notCircle:  # arc
                    if (
                        separator > 0 and shifts.index(" ") > 0
                    ):  # some cases have "a 0,0 0 0 1 0,0"
                        continue
                    point_seq.append([float(af[-2]), float(af[-1])])
                elif char == "a" and not notCircle:
                    if (
                        separator > 0 and shifts.index(" ") > 0
                    ):  # some cases have "a 0,0 0 0 1 0,0"
                        continue
                    point_seq.append(
                        [previous_x + float(af[-2]), previous_y + float(af[-1])]
                    )
                elif char == "M" or char == "L":  # move to or line to
                    point_seq.append([float(af[0]), float(af[1])])
                    af = af[2:]
                    if char == "L" and len(af) % 2 == 0 and len(af) > 0:
                        while len(af) > 0:
                            point_seq.append([float(af[0]), float(af[1])])
                            af = af[2:]
                elif char == "m" or char == "l":
                    point_seq.append(
                        [previous_x + float(af[0]), previous_y + float(af[1])]
                    )
                elif char == "C":  # cubic-bezier
                    point_seq.append([float(af[-2]), float(af[-1])])
                elif char == "c":
                    point_seq.append(
                        [previous_x + float(af[-2]), float(previous_y + af[-1])]
                    )
                elif char == "V":  # vertical line
                    point_seq.append([previous_x, float(shifts)])
                elif char == "v":
                    # if (parseFloat(shifts) !== 0)
                    # tbd: here we cannot use this if otherwise 0hw-rect will be filtered out;
                    point_seq.append([previous_x, previous_y + float(shifts)])
                elif char == "H":  # horizontal line
                    point_seq.append([float(shifts), previous_y])
                elif char == "h":
                    point_seq.append([previous_x + float(shifts), previous_y])

    # then check whether the point seq forms a rect or a line
    # haven't deal with all possible cases
    if (
        len(point_seq) == 5
        and point_seq[0][0] == point_seq[-1][0]
        and point_seq[0][1] == point_seq[-1][1]
    ) and (char_seq[0:4] == "MLLL" or char_seq[0:4] == "Mlll"):
        width = (
            max(point_seq, key=lambda p: p[0])[0]
            - min(point_seq, key=lambda p: p[0])[0]
        )
        height = (
            max(point_seq, key=lambda p: p[1])[1]
            - min(point_seq, key=lambda p: p[1])[1]
        )
        spec["tag"] = "rect"
        spec["x"] = min(point_seq, key=lambda p: p[0])[0]
        spec["y"] = min(point_seq, key=lambda p: p[1])[1]
        spec["width"] = width
        spec["height"] = height
    elif len(point_seq) == 2:
        spec["tag"] = "line"
        spec["x1"] = point_seq[0][0]
        spec["y1"] = point_seq[0][1]
        spec["x2"] = point_seq[1][0]
        spec["y2"] = point_seq[1][1]
    elif len(point_seq) == 3 and (char_seq == "MAA" or char_seq == "Maa"):
        if point_seq[0][0] == point_seq[2][0] and point_seq[0][1] == point_seq[2][1]:
            spec["tag"] = "circle"
            spec["cx"] = 0.5 * (point_seq[0][0] + point_seq[1][0])
            spec["cy"] = 0.5 * (point_seq[0][1] + point_seq[1][1])
            # todo: radius
            spec["r"] = 0.5 * (point_seq[0][0] - point_seq[1][0])
    elif re.match(r"^M(C|c){8}(Z|z)$", char_seq):
        if (
            abs(
                abs(point_seq[0][0] - point_seq[2][0])
                - abs(point_seq[0][1] - point_seq[2][1])
            )
            < PRECISION
        ):
            spec["tag"] = "circle"
            spec["r"] = abs(point_seq[0][0] - point_seq[2][0])
            spec["cx"] = 0.5 * (point_seq[0][0] + point_seq[4][0])
            spec["cy"] = 0.5 * (point_seq[0][1] + point_seq[4][1])
    elif re.match(r"^M(C|c)+(L|l|M)z$", char_seq):
        if len(point_seq) % 2 == 1:
            mid = int((len(point_seq) - 1) / 2) - 1
            # (-2,mid) (0,-3) 90 degree
            if (
                abs(
                    (point_seq[0][0] - point_seq[-3][0])
                    * (point_seq[mid][0] - point_seq[-2][0])
                    + (point_seq[0][1] - point_seq[-3][1])
                    * (point_seq[mid][1] - point_seq[-2][1])
                )
                < 0.001
            ):
                spec["tag"] = "arc"
                spec["cx"] = point_seq[-2][0]
                spec["cy"] = point_seq[-2][1]
                spec["arc"] = [point_seq[0], point_seq[mid], point_seq[-3]]
                spec["r"] = (
                    (point_seq[0][0] - point_seq[-2][0]) ** 2
                    + (point_seq[0][1] - point_seq[-2][1]) ** 2
                ) ** 0.5
                cos = (
                    (point_seq[0][0] - point_seq[-2][0])
                    * (point_seq[-3][0] - point_seq[-2][0])
                    + (point_seq[0][1] - point_seq[-2][1])
                    * (point_seq[-3][1] - point_seq[-2][1])
                ) / (spec["r"] ** 2)
                # avoid precision error, e.g., -1.0000000000000002
                cos = max(cos, -1)
                cos = min(cos, 1)
                spec["theta"] = math.acos(cos)
                if len(point_seq) > 7:
                    spec["theta"] = 2 * math.pi - spec["theta"]
    # below is for handling cases where using a rect-like shapes to represent axis lines
    elif len(point_seq) > 1:
        width = (
            max(point_seq, key=lambda p: p[0])[0]
            - min(point_seq, key=lambda p: p[0])[0]
        )
        height = (
            max(point_seq, key=lambda p: p[1])[1]
            - min(point_seq, key=lambda p: p[1])[1]
        )
        if (
            width < 2
            and np.max(np.abs(np.diff(np.array([p[1] for p in point_seq])))) < 20
        ):
            spec["tag"] = "line"
            spec["x1"] = max(point_seq, key=lambda p: p[0])[0]
            spec["y1"] = min(point_seq, key=lambda p: p[1])[1]
            spec["x2"] = max(point_seq, key=lambda p: p[0])[0]
            spec["y2"] = max(point_seq, key=lambda p: p[1])[1]
        elif (
            height < 2
            and np.max(np.abs(np.diff(np.array([p[0] for p in point_seq])))) < 20
        ):
            spec["tag"] = "line"
            spec["x1"] = min(point_seq, key=lambda p: p[0])[0]
            spec["y1"] = min(point_seq, key=lambda p: p[1])[1]
            spec["x2"] = max(point_seq, key=lambda p: p[0])[0]
            spec["y2"] = min(point_seq, key=lambda p: p[1])[1]
        elif re.match(r"^(M|m|L|l)+", char_seq):
            spec["tag"] = "path"
            spec["points"] = point_seq
    if spec["tag"] == "path":
        spec["d"] = draw_seq
    return spec


def process_axis_matplotlib(spec):
    if "children" not in spec or len(spec["children"]) == 0:
        return
    spec["type"] = "axis"
    spec["tick"] = []
    spec["tick_value"] = []
    for child in spec["children"]:
        child_id = child["id"]
        if "xtick" in child_id:
            spec["type"] = "xaxis"
        elif "ytick" in child_id:
            spec["type"] = "yaxis"

        if "tick" in child_id:
            has_tick_line = any(item["tag"] == "line" for item in child["children"])
            for child2 in child["children"]:
                if child2["tag"] == "line":
                    if spec["type"] == "xaxis":
                        if (
                            len(spec["tick_value"]) == 0
                            or abs(child2["x1"] - spec["tick_value"][-1]) > PRECISION
                        ):
                            spec["tick_value"].append(child2["x1"])
                    elif spec["type"] == "yaxis":
                        if (
                            len(spec["tick_value"]) == 0
                            or abs(child2["y1"] - spec["tick_value"][-1]) > PRECISION
                        ):
                            spec["tick_value"].append(child2["y1"])
                if child2["tag"] == "text":
                    spec["tick"].append(child2["content"])
                    if not has_tick_line:
                        if spec["type"] == "xaxis":
                            spec["tick_value"].append(child2["x"] + child2["width"] / 2)
                        elif spec["type"] == "yaxis":
                            spec["tick_value"].append(child2["y"])

        elif "text" in child_id:
            if re.match(r"^1e[1-9][0-9]?$", child["content"]):
                spec["unit"] = eval(child["content"])
            else:
                spec["title"] = child["content"]


def process_legend_matplotlib(spec):
    spec["type"] = "legend"
    labels = []
    examples = []
    # element(index 0) might be background
    # todo: A more accurate way to recognize background
    first = 0 if spec["children"][0]["tag"] == "text" else 1
    for i in range(first, len(spec["children"])):
        child = spec["children"][i]
        if child["tag"] == "text":
            labels.append(child)
        else:
            examples.append(child)
    # circle > line
    circles = list(filter(lambda item: item["tag"] == "circle", examples))
    if len(circles) > 0:
        examples = list(filter(lambda item: item["tag"] != "line", examples))
    # rect > line
    rects = list(filter(lambda item: item["tag"] == "rect", examples))
    if len(rects) > 0:
        examples = list(filter(lambda item: item["tag"] != "line", examples))
    if len(examples) < 2:
        return
    # find the most common tag
    property_counts = Counter(item["tag"] for item in examples)
    most_common_tag = property_counts.most_common(1)
    if most_common_tag[0][1] < 2:
        return
    examples = list(filter(lambda item: item["tag"] == most_common_tag[0][0], examples))

    # if example is group, combination
    if len(examples) > 0 and examples[0]["tag"] == "g":
        # hypothesis: line > circle
        children_tags = [child["tag"] for child in examples[0]["children"]]
        if "line" in children_tags:
            child_tag = "line"
        new_examples = []
        for example in examples:
            for child in example["children"]:
                if child["tag"] == child_tag:
                    new_examples.append(child)
                    break
        if len(new_examples) == len(examples):
            examples = new_examples

    differences = []
    for channel in channels:
        if channel in examples[0]:
            value = examples[0][channel]
            is_str = isinstance(value, str)
            for i in range(1, len(examples)):
                if (
                    channel not in examples[i]
                    or (is_str and examples[i][channel] != value)
                    or (not is_str and abs(examples[i][channel] - value) > PRECISION)
                ):
                    differences.append(channel)
                    break

    # # hypothesis: legend label in same row or column
    # todo: bug: 1 title + 1 label
    if len(labels) > len(examples) and len(labels) > 2:
        labels_x = [label["x"] for label in labels]
        labels_y = [label["y"] for label in labels]
        for i in range(len(labels) - 1, -1, -1):
            if labels_x.count(labels_x[i]) == 1 and labels_y.count(labels_y[i]) == 1:
                # maybe title
                spec["title"] = labels[i]["content"]
                del labels[i]

    dists = []
    for i in range(len(examples)):
        if examples[i]["tag"] == "rect":
            dists = [
                (label["x"] - examples[i]["x"]) ** 2
                + (label["y"] - examples[i]["y"]) ** 2
                for label in labels
            ]
        elif examples[i]["tag"] == "line":
            dists = [
                (label["x"] - examples[i]["x1"]) ** 2
                + (label["y"] - examples[i]["y1"]) ** 2
                for label in labels
            ]
        elif examples[i]["tag"] == "circle":
            dists = [
                (label["x"] - examples[i]["cx"]) ** 2
                + (label["y"] - examples[i]["cy"]) ** 2
                for label in labels
            ]

        if len(dists) > 0:
            min_index = dists.index(min(dists))
            examples[i]["label"] = labels[min_index]["content"]

    mapping = {}
    # remove redundancy
    if "fill" in differences and "stroke" in differences:
        flag = True
        for example in examples:
            if (
                "stroke" in example
                and "fill" not in example
                and example["stroke"] == "#000000"
            ):
                example["fill"] = "#000000"

            if example["stroke"] != example["fill"]:
                flag = False
                break
        if flag:
            differences.remove("stroke")

    for example in examples:
        key = ""
        for difference in differences:
            key += difference + ":" + str(example[difference]) + ";"
        if key in mapping:
            mapping = {}
            return False, f"(legend) same style for different labels: {key}"
        mapping[key] = example["label"]
    spec["mapping"] = mapping
    spec["channel"] = differences


def process_transform(node, spec, acc_trans):
    trans_str = get_attribute_value(node, "transform")
    if "translate" in trans_str:
        index_of_translate = trans_str.index("translate")
        new_str = trans_str[index_of_translate + 9 :]
        index_of_right_bracket = new_str.index(")")
        index_of_left_bracket = new_str.index("(")
        value = new_str[index_of_left_bracket + 1 : index_of_right_bracket].strip()
        if "," in value:
            dxy = value.split(",")
            dx = float(dxy[0])
            dy = float(dxy[1])
        else:
            dxy = value.split(" ")
            dx = float(dxy[0])
            if len(dxy) == 1:
                dy = 0.0
            else:
                dy = float(dxy[-1])
        spec["translate"] = [dx + acc_trans[0], dy + acc_trans[1]]
    if "matrix" in trans_str:
        index_of_matrix = trans_str.index("matrix")
        new_str = trans_str[index_of_matrix + 6 :]
        index_of_right_bracket = new_str.index(")")
        index_of_left_bracket = new_str.index("(")
        value = new_str[index_of_left_bracket + 1 : index_of_right_bracket].strip()
        if "," in value:
            matrix = value.split(",")
        else:
            matrix = value.split(" ")
        if "matrix" not in spec:
            spec["matrix"] = matrix
        else:
            # todo: concatenate matrix
            spec["matrix"] = matrix + spec["matrix"]
    if "scale" in trans_str:
        index_of_translate = trans_str.index("scale")
        new_str = trans_str[index_of_translate + 5 :]
        index_of_right_bracket = new_str.index(")")
        index_of_left_bracket = new_str.index("(")
        value = new_str[index_of_left_bracket + 1 : index_of_right_bracket].strip()
        if "," in value:
            scale_xy = value.split(",")
            scale_x = float(scale_xy[0])
            scale_y = float(scale_xy[1])
        else:
            scale_xy = value.split(" ")
            scale_x = float(scale_xy[0])
            if len(scale_xy) == 1:
                scale_y = scale_x
            else:
                scale_y = float(scale_xy[1])
        spec["scale"] = [abs(scale_x), abs(scale_y)]
    # if 'rotate' in trans_str:
    #     spec['rotate'] = True
    # if 'scale' in trans_str:
    #     spec['scale'] = True
    # if 'skewX' in trans_str:
    #     spec['skewX'] = True
    # if 'skewY' in trans_str:
    #     spec['skewY'] = True


def extract_features(node, spec, parent, acc_trans, global_style):
    # analysis attributes
    attrs = get_attribute_names(node)
    if spec["tag"] == "style":
        pass
        # todo: class or tag based style

    # process path
    if spec["tag"] == "path":
        process_path(node, spec)

    # todo: process polyline

    # inheriting parent's attributes
    if parent and parent["tag"] == "g":
        for key in parent:
            if (
                key in keys_of_interest
                and key != "id"
                and key != "width"
                and key != "height"
            ):
                spec[key] = parent[key]

    for attr in keys_of_interest:
        if attr in attrs:
            if attr == "style":
                local_styles = get_attribute_value(node, attr).split(";")
                for pair in local_styles:
                    if pair == "":
                        continue
                    pair = pair.strip().split(":")
                    if len(pair) == 2:
                        if pair[0].strip() in keys_of_interest:
                            spec[pair[0].strip()] = pair[1].strip()
            elif attr == "transform":
                # process transform
                process_transform(node, spec, acc_trans)
            elif attr == "font-size":
                pass
            elif attr == "dx" or attr == "dy":
                pass
            elif attr == "x" or attr == "y":
                spec[attr] = float(get_attribute_value(node, attr))
            else:
                spec[attr] = get_attribute_value(node, attr).strip()

    if "fill-opacity" in spec:
        if spec["fill-opacity"] == "0" and "fill" in parent:
            spec["fill"] = parent["fill"]
            del spec["fill-opacity"]

    if (
        ("transform" not in attr or get_attribute_value(node, "transform") == "")
        and acc_trans[0] != 0
        and acc_trans[1] != 0
    ):
        spec["transform"] = f"translate({acc_trans[0]}, {acc_trans[1]})"


# node: element node
# acc_trans: accumulated transformation
# global_style: global style
def parser_node(
    node, parent, defss={}, acc_trans=[0, 0], global_style={}, source="matplotlib"
):
    spec = {}
    spec["children"] = []

    tag = extract_tag(node)
    spec["tag"] = tag

    if tag == "metadata":
        return spec

    if tag == "style":
        # todo: global style
        pass

    extract_features(node, spec, parent, acc_trans, global_style)

    if tag == "defs":
        # todo: defs
        for child in node.childNodes:
            if child.nodeType == child.ELEMENT_NODE:
                child_spec = parser_node(child, spec, defss)
                if "id" in child_spec:
                    defss[child_spec["id"]] = child_spec
        return spec

    if tag == "use":
        defs = defss[get_attribute_value(node, "xlink:href")[1:]]
        if defs:  # copy
            spec["tag"] = defs["tag"]
            for key in defs:
                if key != "id" and key not in spec:
                    spec[key] = defs[key]
        # todo: transform
        if "x" in spec:
            if spec["tag"] == "line":
                spec["x1"] += spec["x"]
                spec["x2"] += spec["x"]
                del spec["x"]
            elif spec["tag"] == "circle":
                spec["cx"] += spec["x"]
                del spec["x"]
        if "y" in spec:
            if spec["tag"] == "line":
                spec["y1"] += spec["y"]
                spec["y2"] += spec["y"]
                del spec["y"]
            elif spec["tag"] == "circle":
                spec["cy"] += spec["y"]
                del spec["y"]

    # process text
    if tag == "g" and source == "matplotlib":
        node_id = node.getAttribute("id")
        if "text" in node_id:
            process_text_matplotlib(node, spec, acc_trans)

    if node.childNodes.length > 0 and spec["tag"] != "text":
        for child in node.childNodes:
            # Pass Text node \n
            if child.nodeType == child.ELEMENT_NODE:
                parser_node(child, spec, defss, acc_trans, global_style, source)
        # remove useless children
        if len(spec["children"]) == 1 and spec["tag"] == "g":
            if "id" not in spec["children"][0]:
                if "id" in spec:
                    spec["children"][0]["id"] = spec["id"]
                spec = spec["children"][0]

    # process matplotlib
    if spec["tag"] == "g" and source == "matplotlib" and "type" not in spec:
        node_id = node.getAttribute("id")
        # remove background
        if "figure" in node_id:
            spec["type"] = "figure"
            if spec["children"][0]["tag"] == "rect":
                spec["width"] = spec["children"][0]["width"]
                spec["height"] = spec["children"][0]["height"]
                del spec["children"][0]
        elif "axes" in node_id:
            spec["type"] = "subplot"
            # remove background
            if (
                spec["children"][0]["tag"] == "rect"
                and "fill" in spec["children"][0]
                and spec["children"][0]["fill"] == "#ffffff"
            ):
                spec["width"] = spec["children"][0]["width"]
                spec["height"] = spec["children"][0]["height"]
                spec["x"] = spec["children"][0]["x"]
                spec["y"] = spec["children"][0]["y"]
                del spec["children"][0]
                # remove axis lines
                for index in range(len(spec["children"]) - 2, -1, -1):
                    child = spec["children"][index]
                    if child["tag"] == "line":
                        if (
                            abs(child["x1"] - child["x2"]) == spec["width"]
                            and child["y1"] == child["y2"]
                        ):
                            del spec["children"][index]
                        elif (
                            abs(child["y1"] - child["y2"]) == spec["height"]
                            and child["x1"] == child["x2"]
                        ):
                            del spec["children"][index]
        elif "matplotlib.axis" in node_id:
            process_axis_matplotlib(spec)
        elif "legend" in node_id:
            process_legend_matplotlib(spec)

    if "children" in spec and len(spec["children"]) == 0:
        del spec["children"]

    if parent is not None:
        if spec["tag"] == "g" and (
            "children" not in spec or len(spec["children"]) == 0
        ):
            # remove empty group
            pass
        elif spec["tag"] == "path" and spec["d"] == "":
            # remove path without "d"
            pass
        else:
            parent["children"].append(spec)
    return spec


def is_numeric(s):
    try:
        float(s)
        return True
    except ValueError:
        try:
            float(s.replace("−", "-"))
            return True
        except ValueError:
            return False


def analysis_data_type(data_domain, data_range):
    if len(data_domain) <= 1:
        return "nominal"

    # range not number -> nominal
    numeric_values = [is_numeric(s) for s in data_range]
    if False in numeric_values:
        return "nominal"

    if len(data_domain) == len(set(data_domain)):
        numeric_values = [is_numeric(s) for s in data_domain]
        if False not in numeric_values:
            return "quantitative"

        datetime_values = [is_datetime(s) for s in data_domain]
        if False not in datetime_values:
            return "temporal"
    return "nominal"


def analysis_axis(spec, encoding):
    if spec["type"] == "xaxis":
        channel = "x"
    elif spec["type"] == "yaxis":
        channel = "y"

    if channel:
        encoding[channel] = {
            "field": "field_" + channel,
            "type": analysis_data_type(spec["tick"], spec["tick_value"]),
        }
        if len(spec["tick"]) == len(spec["tick_value"]):
            encoding[channel]["scale"] = {
                "domain": spec["tick"],
                "range": spec["tick_value"],
            }
        else:
            # todo
            if len(spec["tick"]) > 1:
                gap = (len(spec["tick_value"]) - len(spec["tick"])) / (
                    len(spec["tick"]) - 1
                )
            else:
                gap = 0
            if int(gap) == gap:
                gap = int(gap)
                scale_range = [
                    spec["tick_value"][index + (index - 1) * gap]
                    for index in range(len(spec["tick"]))
                ]
                encoding[channel]["scale"] = {
                    "domain": spec["tick"],
                    "range": scale_range,
                }
            else:
                encoding[channel]["scale"] = {
                    "domain": spec["tick"],
                    "range": spec["tick_value"],
                }
        if "unit" in spec:
            encoding[channel]["unit"] = spec["unit"]
        if "title" in spec:
            encoding[channel]["title"] = spec["title"]


def analysis_legend(spec, encoding):
    if "channel" in spec and len(spec["channel"]) == 1:
        channel = spec["channel"][0]
        data_domain = []
        data_range = []
        for key in spec["mapping"]:
            data_domain.append(spec["mapping"][key])
            data_range.append(key.split(";")[0].split(":")[1])
        encoding[channel] = {
            "field": "field_" + channel,
            "scale": {
                "domain": data_domain,
                "range": data_range,
            },
            "type": analysis_data_type(data_domain, data_range),
        }
        if "title" in spec:
            encoding[channel]["title"] = spec["title"]


def get_aligned_index(rect, channel, ticks):
    if channel == "x":
        mid = rect["x"] + 0.5 * rect["width"]
    elif channel == "y":
        mid = rect["y"] + 0.5 * rect["height"]

    if mid is not None:
        if len(ticks) == 1:
            if abs(mid - ticks[0]) < PRECISION:
                return 0
        else:
            tick_width = (max(ticks) - min(ticks)) / (len(ticks) - 1)
            minus = np.array([abs(mid - tick) for tick in ticks])
            indexes = np.where(minus < (PRECISION + tick_width / 2))

            if len(indexes[0]) > 0:
                return indexes[0][0]
    return None


def is_align_with_axis(rect, channel, ticks):
    # todo: handle ticks without tick line
    if len(ticks) == 0:
        return True
    if get_aligned_index(rect, channel, ticks) is not None:
        return True
    return False


def identify_mark_arcs(arcs):
    filtered_arcs = arcs
    # filter shadow
    filtered_arcs = list(
        filter(
            lambda arc: "opacity" not in arc or float(arc["opacity"]) > 0.6,
            filtered_arcs,
        )
    )
    return filtered_arcs


def identify_mark_circles(circles, spec):
    encoding = spec["encoding"]
    filtered_circles = circles

    if len(circles) > 1:
        radius = [circle["r"] for circle in circles]
        if max(radius) - min(radius) < PRECISION:
            return circles

    if "x" in encoding and "y" in encoding:
        if (
            encoding["x"]["type"] == "quantitative"
            and encoding["y"]["type"] == "quantitative"
        ):
            # in range
            x_range = encoding["x"]["scale"]["range"]
            y_range = encoding["y"]["scale"]["range"]
            filtered_circles = list(
                filter(
                    lambda circle: circle["cx"] >= x_range[0]
                    and circle["cx"] <= x_range[1]
                    and circle["cy"] >= y_range[1]
                    and circle["cy"] <= y_range[0],
                    filtered_circles,
                )
            )

        for channel in encoding:
            if channel != "x" and channel != "y":
                if encoding[channel]["type"] == "nominal":
                    filtered_circles = list(
                        filter(
                            lambda circle: circle[channel]
                            in encoding[channel]["scale"]["range"],
                            filtered_circles,
                        )
                    )

    return filtered_circles


def identify_mark_bars(rects, spec):
    encoding = spec["encoding"]

    if "x" in encoding and "y" in encoding:
        x_range = encoding["x"]["scale"]["range"]
        y_range = encoding["y"]["scale"]["range"]

        if len(rects) > 1:
            width = [rect["width"] for rect in rects]
            # vertical bar? same width ?
            if max(width) - min(width) < PRECISION:
                return rects
            # ## horizontal bar?
            height = [rect["height"] for rect in rects]
            # same height ?
            if max(height) - min(height) < PRECISION:
                return rects

        # y: in range
        filtered_rects = rects
        if len(y_range) > 1:
            filtered_rects = list(
                filter(
                    lambda rect: rect["y"] >= min(y_range[0], y_range[1])
                    and rect["y"] + rect["height"] <= max(y_range[0], y_range[1]),
                    filtered_rects,
                )
            )
        if len(filtered_rects) > 0:
            width = [rect["width"] for rect in filtered_rects]
            # same width ?
            if max(width) - min(width) > PRECISION:
                filtered_rects = []

        if len(filtered_rects) == 0:
            # x: in range
            filtered_rects = rects
            if len(x_range) > 1:
                filtered_rects = list(
                    filter(
                        lambda rect: rect["x"] >= min(x_range[0], x_range[1])
                        and rect["x"] + rect["width"] <= max(x_range[0], x_range[1]),
                        filtered_rects,
                    )
                )
            if len(filtered_rects) > 0:
                height = [rect["height"] for rect in filtered_rects]
                # same height ?
                if max(height) - min(height) > PRECISION:
                    return []

    for channel in encoding:
        if channel != "x" and channel != "y":
            if encoding[channel]["type"] == "nominal":
                filtered_rects = list(
                    filter(
                        lambda rect: rect[channel]
                        in encoding[channel]["scale"]["range"],
                        filtered_rects,
                    )
                )

    return filtered_rects


def identify_mark_lines(paths):
    filtered_paths = paths
    filtered_paths = list(filter(lambda path: "points" in path, filtered_paths))
    return filtered_paths


def analysis_mark(nodes, spec):
    encoding = spec["encoding"]
    spec["data"] = []
    if "arc" in nodes:
        arcs = nodes["arc"]
        arcs = identify_mark_arcs(arcs)
        if len(arcs) > 0:
            # add encoding theta
            encoding["theta"] = {
                "field": "field_theta",
                "type": "quantitative",
                "scale": {"domain": [0, 100], "range": [0, 2 * math.pi]},  # percent
            }
            if "fill" not in encoding:
                data_domain = []
                data_range = []
                # find label
                # if no label -> error
                if len(nodes["text"]) >= len(arcs):
                    for arc in arcs:
                        data_range.append(arc["fill"])
                        # value or label?
                        dists = []
                        # filter out value
                        nodes["text"] = [
                            text for text in nodes["text"] if "%" not in text["content"]
                        ]
                        # filter null string
                        nodes["text"] = [
                            text
                            for text in nodes["text"]
                            if len(text["content"].strip()) > 0
                        ]
                        for text in nodes["text"]:
                            # label or title
                            arc_x = arc["arc"][1][0]
                            arc_y = arc["arc"][1][1]
                            if arc_x >= arc["cx"]:
                                # right
                                dists.append(
                                    (arc_y - text["y"]) ** 2 + (arc_x - text["x"]) ** 2
                                )
                            else:
                                # left
                                dists.append(
                                    (arc_y - text["y"]) ** 2
                                    + (arc_x - text["x"] - text["width"]) ** 2
                                )
                        min_index = dists.index(min(dists))
                        data_domain.append(nodes["text"][min_index]["content"])
                        nodes["text"].pop(min_index)

                encoding["fill"] = {
                    "field": "field_fill",
                    "type": "nominal",
                    "scale": {"domain": data_domain, "range": data_range},
                }
            spec["mark"] = "arc"
            spec["chart"] = "pie"
            spec["data"] = []
            for arc in arcs:
                arc["type"] = "mark"
                item = {}
                for channel in encoding:
                    data_range = encoding[channel]["scale"]["range"]
                    data_domain = encoding[channel]["scale"]["domain"]
                    if encoding[channel]["type"] == "nominal":
                        index = data_range.index(arc[channel])
                        item[encoding[channel]["field"]] = data_domain[index]
                    elif encoding[channel]["type"] == "quantitative":
                        actual_channel = channel
                        if data_range[1] - data_range[0] != 0:
                            item[encoding[channel]["field"]] = round(
                                data_domain[0]
                                + (data_domain[1] - data_domain[0])
                                / (data_range[1] - data_range[0])
                                * arc[actual_channel],
                                6,
                            )
                        else:
                            item[encoding[channel]["field"]] = data_range[0]

                spec["data"].append(item)

    if ("rect" in nodes) and ("mark" not in spec):
        rects = nodes["rect"]
        # bar chart
        bars = identify_mark_bars(rects, spec)
        if len(bars) > 0:
            spec["mark"] = "bar"
            width = [bar["width"] for bar in bars]
            if len(bars) == 1:
                if encoding["x"]["type"] != "quantitative":
                    spec["chart"] = "vertical "
                else:
                    spec["chart"] = "horizontal "
            else:
                if max(width) - min(width) < PRECISION:
                    spec["chart"] = "vertical "
                else:
                    spec["chart"] = "horizontal "

            spec["data"] = []
            for bar in bars:
                bar["type"] = "mark"
                item = {}
                for channel in encoding.copy():
                    data_range = encoding[channel]["scale"]["range"]
                    data_domain = encoding[channel]["scale"]["domain"]
                    if encoding[channel]["type"] == "nominal":
                        if channel == "x" or channel == "y":
                            # todo: Get index in order
                            if len(data_domain) == len(bars):
                                index = bars.index(bar)
                            else:
                                index = get_aligned_index(bar, channel, data_range)
                            item[encoding[channel]["field"]] = data_domain[index]
                        else:
                            try:
                                index = data_range.index(bar[channel])
                                item[encoding[channel]["field"]] = data_domain[index]
                            except ValueError:
                                del encoding[channel]
                                continue
                    elif (
                        encoding[channel]["type"] == "quantitative"
                        or encoding[channel]["type"] == "temporal"
                    ):
                        invert = 1
                        if channel == "y":
                            if "horizontal" in spec["chart"]:
                                value = bar["y"] + bar["height"] / 2 - data_range[0]
                            else:
                                # zero point in left top
                                if data_range[1] < data_range[0]:
                                    invert = -1
                                value = bar["height"]
                        elif channel == "x":
                            if "horizontal" in spec["chart"]:
                                value = bar["width"]
                            else:
                                value = bar["x"] + bar["width"] / 2 - data_range[0]
                        if (
                            encoding[channel]["type"] == "quantitative"
                            and data_domain[1] * data_domain[0] < 0
                            and data_domain[0] < -PRECISION
                        ):
                            # in case rang [-, +]
                            zero_point = (
                                data_domain[1] * data_range[0]
                                - data_domain[0] * data_range[1]
                            ) / (data_domain[1] - data_domain[0])
                            # positive
                            if abs((bar[channel] - zero_point)) > PRECISION:
                                item[encoding[channel]["field"]] = round(
                                    (data_domain[1] - data_domain[0])
                                    / (data_range[1] - data_range[0])
                                    * (bar[channel] - zero_point),
                                    6,
                                )
                            else:  # negative
                                item[encoding[channel]["field"]] = round(
                                    invert
                                    * (data_domain[1] - data_domain[0])
                                    / (data_range[1] - data_range[0])
                                    * (-value),
                                    6,
                                )
                        elif data_range[1] - data_range[0] != 0:
                            item[encoding[channel]["field"]] = round(
                                data_domain[0]
                                + invert
                                * (data_domain[1] - data_domain[0])
                                / (data_range[1] - data_range[0])
                                * (value),
                                6,
                            )
                        else:
                            item[encoding[channel]["field"]] = data_range[0]

                        if encoding[channel]["type"] == "temporal":
                            item[encoding[channel]["field"] + "_origin"] = item[
                                encoding[channel]["field"]
                            ]
                            item[encoding[channel]["field"]] = parse_timestamp_to_time(
                                item[encoding[channel]["field"]]
                            )
                spec["data"].append(item)
            # identify chart type
            if len(encoding.keys()) > 2:
                x_mins = [bar["x"] for bar in bars]
                y_mins = [bar["y"] for bar in bars]
                x_maxs = [bar["x"] + bar["width"] for bar in bars]
                y_maxs = [bar["y"] + bar["height"] for bar in bars]
                if (
                    max(x_mins) - min(x_mins) < PRECISION
                    or max(x_maxs) - min(x_maxs) < PRECISION
                ):
                    if len(set(y_mins)) == len(y_mins):
                        spec["chart"] += "grouping bar"
                    else:
                        # overlap
                        spec["chart"] += "bar(error)"
                elif (
                    max(y_mins) - min(y_mins) < PRECISION
                    or max(y_maxs) - min(y_maxs) < PRECISION
                ):
                    if len(set(x_mins)) == len(x_mins):
                        spec["chart"] += "grouping bar"
                    else:
                        # overlap
                        spec["chart"] += "bar(error)"
                else:
                    spec["chart"] += "stacked bar"
            else:
                spec["chart"] += "bar"

    if (("path" in nodes) or ("line" in nodes)) and ("mark" not in spec):
        lines = []
        if "path" in nodes:
            paths = nodes["path"]
            lines += identify_mark_lines(paths)
        if "line" in nodes:
            for line in nodes["line"]:
                if (
                    abs(line["x1"] - line["x2"]) > PRECISION
                    and abs(line["y1"] - line["y2"]) > PRECISION
                ):
                    line["points"] = [
                        [line["x1"], line["y1"]],
                        [line["x2"], line["y2"]],
                    ]
                    lines.append(line)

        if len(lines) > 0:
            spec["mark"] = "line"
            if len(encoding.keys()) > 2:
                spec["chart"] = "grouping line"
            else:
                spec["chart"] = "line"
            spec["data"] = []
            for line in lines:
                line["type"] = "mark"
                item = {}
                for channel in encoding:
                    if channel != "x" and channel != "y":
                        data_range = encoding[channel]["scale"]["range"]
                        data_domain = encoding[channel]["scale"]["domain"]
                        if encoding[channel]["type"] == "nominal":
                            index = data_range.index(line[channel])
                            item[encoding[channel]["field"]] = data_domain[index]
                        elif encoding[channel]["type"] == "quantitative":
                            actual_channel = channel
                            if data_range[1] - data_range[0] != 0:
                                item[encoding[channel]["field"]] = round(
                                    data_domain[0]
                                    + (data_domain[1] - data_domain[0])
                                    / (data_range[1] - data_range[0])
                                    * (line[actual_channel] - data_range[0]),
                                    6,
                                )
                            else:
                                item[encoding[channel]["field"]] = data_range[0]
                for point in line["points"]:
                    item1 = item.copy()
                    for channel in ["x", "y"]:
                        data_range = encoding[channel]["scale"]["range"]
                        data_domain = encoding[channel]["scale"]["domain"]
                        if channel == "x":
                            value = point[0]
                        else:
                            value = point[1]
                        if encoding[channel]["type"] == "nominal":
                            try:
                                index = data_range.index(value)
                                item1[encoding[channel]["field"]] = data_domain[index]
                            except ValueError:
                                # cannot understand the chart
                                spec["data"] = []
                                return
                            except IndexError:
                                spec["data"] = []
                                return
                        elif (
                            encoding[channel]["type"] == "quantitative"
                            or encoding[channel]["type"] == "temporal"
                        ):
                            if (data_range[1] - data_range[0]) != 0:
                                item1[encoding[channel]["field"]] = round(
                                    data_domain[0]
                                    + (data_domain[1] - data_domain[0])
                                    / (data_range[1] - data_range[0])
                                    * (value - data_range[0]),
                                    6,
                                )
                            else:
                                item1[encoding[channel]["field"]] = data_range[0]
                            if encoding[channel]["type"] == "temporal":
                                item1[encoding[channel]["field"] + "_origin"] = item1[
                                    encoding[channel]["field"]
                                ]
                                item1[
                                    encoding[channel]["field"]
                                ] = parse_timestamp_to_time(
                                    item1[encoding[channel]["field"]]
                                )

                    spec["data"].append(item1)

    if ("circle" in nodes) and ("mark" not in spec):
        circles = nodes["circle"]
        # scatter plot
        circles = identify_mark_circles(circles, spec)
        if len(circles) > 0:
            spec["mark"] = "circle"
            spec["data"] = []
            for circle in circles:
                circle["type"] = "mark"
                item = {}
                for channel in encoding.copy():
                    actual_channel = channel
                    if channel == "y":
                        actual_channel = "cy"
                    elif channel == "x":
                        actual_channel = "cx"
                    data_range = encoding[channel]["scale"]["range"]
                    data_domain = encoding[channel]["scale"]["domain"]
                    if encoding[channel]["type"] == "nominal":
                        try:
                            index = data_range.index(circle[actual_channel])
                        except ValueError:
                            if channel != "x" and channel != "y":
                                del encoding[channel]
                                continue
                        item[encoding[channel]["field"]] = data_domain[index]
                    elif (
                        encoding[channel]["type"] == "quantitative"
                        or encoding[channel]["type"] == "temporal"
                    ):
                        if data_range[1] - data_range[0] != 0:
                            item[encoding[channel]["field"]] = round(
                                data_domain[0]
                                + (data_domain[1] - data_domain[0])
                                / (data_range[1] - data_range[0])
                                * (circle[actual_channel] - data_range[0]),
                                6,
                            )
                        else:
                            item[encoding[channel]["field"]] = data_range[0]

                        if encoding[channel]["type"] == "temporal":
                            item[encoding[channel]["field"] + "_origin"] = item[
                                encoding[channel]["field"]
                            ]
                            item[encoding[channel]["field"]] = parse_timestamp_to_time(
                                item[encoding[channel]["field"]]
                            )
                spec["data"].append(item)
                if len(encoding.keys()) > 2:
                    spec["chart"] = "grouping scatter"
                else:
                    spec["chart"] = "scatter"

    if "text" in nodes:
        if len(nodes["text"]) == 1:
            spec["title"] = nodes["text"][0]["content"]


def analysis_scale(spec):
    encoding = spec["encoding"]
    for channel in encoding:
        scale = encoding[channel]["scale"]
        scale["ticks"] = scale["domain"]
        if "quantitative" == encoding[channel]["type"]:
            # error case, no tick value in axis
            if len(scale["domain"]) < 1:
                encoding[channel]["type"] = "nominal"
            scale["domain"] = [
                float(value.replace("−", "-")) for value in scale["domain"]
            ]
            if "unit" in encoding[channel]:
                scale["domain"] = [
                    value * encoding[channel]["unit"] for value in scale["domain"]
                ]
                del encoding[channel]["unit"]
        elif "temporal" == encoding[channel]["type"]:
            scale["domain"] = [
                parse_time_to_timestamp(value) for value in scale["domain"]
            ]
        if "nominal" != encoding[channel]["type"]:
            linear = True
            if len(scale["domain"]) != len(scale["range"]):
                linear = False
            else:
                for i in range(2, len(scale["range"])):
                    k1 = (scale["domain"][i] - scale["domain"][i - 1]) / (
                        scale["range"][i] - scale["range"][i - 1]
                    )
                    k2 = (scale["domain"][i - 1] - scale["domain"][i - 2]) / (
                        scale["range"][i - 1] - scale["range"][i - 2]
                    )
                    if abs((k1 - k2) / k1) > PRECISION:
                        linear = False
                        break

            if linear:
                scale["type"] = "linear"
                if len(scale["domain"]) > 1:
                    k = (scale["range"][-1] - scale["range"][0]) / (
                        scale["domain"][-1] - scale["domain"][0]
                    )
                    b = scale["range"][0] - k * scale["domain"][0]

                    if channel == "x":
                        if "x" in spec and "width" in spec:
                            scale["range"] = [spec["x"], spec["x"] + spec["width"]]
                            scale["domain"] = [
                                (scale["range"][0] - b) / k,
                                (scale["range"][1] - b) / k,
                            ]
                    elif channel == "y":
                        if "y" in spec and "height" in spec:
                            scale["range"] = [spec["y"] + spec["height"], spec["y"]]
                            scale["domain"] = [
                                (scale["range"][0] - b) / k,
                                (scale["range"][1] - b) / k,
                            ]
                max_index = scale["domain"].index(max(scale["domain"]))
                min_index = scale["domain"].index(min(scale["domain"]))
                scale["domain"] = [
                    scale["domain"][min_index],
                    scale["domain"][max_index],
                ]
                scale["range"] = [scale["range"][min_index], scale["range"][max_index]]
            # if not linear scale, maybe nominal
            if "type" not in scale:
                scale["domain"] = scale["ticks"]
                encoding[channel]["type"] = "nominal"


def get_leaf_nodes(spec):
    nodes = []
    if spec["tag"] == "g":
        for child in spec["children"]:
            nodes += get_leaf_nodes(child)
    else:
        nodes.append(spec)
    return nodes


# subplot, error reason
def deconstruct(svg, source="matplotlib"):
    doc = minidom.parseString(svg)
    svg = doc.getElementsByTagName("svg")[0]

    # matplotlib parser
    defss = {}
    spec = parser_node(svg, None, defss, [0, 0], {}, source)
    if (
        "children" not in spec
        or len(spec["children"]) == 0
        or "children" not in spec["children"][0]
    ):
        return None, "Empty figure"

    # Ignore continuous color legend
    subplots = [
        child
        for child in spec["children"][0]["children"]
        if (
            "type" in child
            and child["type"] == "subplot"
            and len(child["children"]) > 0
            and child["children"][0]["tag"] != "image"
        )
    ]

    if len(subplots) == 0:
        return None, "Empty figure"
    elif len(subplots) > 1:
        return None, "Dual axes"

    subplot = subplots[0]

    # find legend
    legends = [
        child
        for child in subplot["children"]
        if ("type" in child and child["type"] == "legend")
    ]
    legend = None
    if len(legends) > 1:
        return None, "Multiple Legend"
    elif len(legends) == 1:
        legend = legends[0]
    else:
        # legend may out of subplot
        legends = [
            child
            for child in spec["children"][0]["children"]
            if ("type" in child and child["type"] == "legend")
        ]
        if len(legends) == 1:
            legend = legends[0]

    subplot["encoding"] = {}
    if legend is not None:
        analysis_legend(legend, subplot["encoding"])

    others = {}
    for child in subplot["children"]:
        if "type" in child:
            if child["type"] == "xaxis" or child["type"] == "yaxis":
                analysis_axis(child, subplot["encoding"])
        else:
            nodes = get_leaf_nodes(child)
            for node in nodes:
                if node["tag"] not in others:
                    others[node["tag"]] = [node]
                else:
                    others[node["tag"]].append(node)
    analysis_scale(subplot)
    analysis_mark(others, subplot)
    return subplot, None



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import base64
import json
import logging
import os
from typing import Union

import cairosvg
import pandas as pd
from attr import dataclass

from .check import (
    chart_check,
    data_check,
    deconstruct,
    layout_check,
    order_check,
    readability_check,
    scale_and_ticks_check,
)
from .dataset import Dataset


@dataclass
class CheckResult:
    answer: Union[bool, int]
    aspect: str
    rationale: str

    def get_json(self):
        return {
            "answer": self.answer,
            "aspect": self.aspect,
            "rationale": self.rationale,
        }


@dataclass
class EvaluationDetail:
    id: str
    results: list[list[CheckResult]]


VALID_ASPECTS = ["code execution", "surface-form check"]
LEGAL_ASPECTS = ["deconstruction", "chart type check", "data check", "order check"]
READABILITY_ASPECT = ["layout check", "scale and ticks check", "readability check"]

FAIL_ASPECTS = VALID_ASPECTS + LEGAL_ASPECTS + ["layout check", "scale and ticks check"]


class EvaluationResult:
    dataset: Dataset
    details: list[EvaluationDetail]

    def __init__(self, dataset: Dataset, details: list[EvaluationDetail]):
        self.dataset = dataset
        self.details = details

    def detail_records(self) -> pd.DataFrame:
        records = []
        for detail in self.details:
            id = detail.id
            instance_results = detail.results
            count = len(instance_results)
            record = {
                "id": id,
                "chart": self.dataset.dict[id]["chart"],
                "hardness": self.dataset.dict[id]["hardness"],
            }

            # fail rate
            for aspect in FAIL_ASPECTS:
                evaluate_result = [
                    (
                        all(
                            [
                                item.answer
                                for item in query_results
                                if item.aspect == aspect
                            ]
                        )
                    )
                    for query_results in instance_results
                ]
                fail_result = [item for item in evaluate_result if not item]
                record[f"{aspect}_fail_rate"] = len(fail_result) / count

            high_level_dimensions = [
                ["invalid_rate", VALID_ASPECTS],
                ["illegal rate", LEGAL_ASPECTS],
            ]
            pass_count = count
            for dimension in high_level_dimensions:
                evaluate_result = [
                    (
                        all(
                            [
                                item.answer
                                for item in query_results
                                if (item.aspect in dimension[1])
                            ]
                        )
                    )
                    for query_results in instance_results
                ]
                false_count = len([item for item in evaluate_result if not item])
                record[dimension[0]] = false_count / count
                pass_count -= false_count
                records.append(record)

            # pass rate
            record["pass_rate"] = pass_count / count
            records.append(record)

            # readability score
            evaluate_result = [
                (
                    sum(
                        [
                            item.answer
                            for item in query_results
                            if item.aspect == "readability check"
                        ]
                    )
                )
                for query_results in instance_results
            ]
            if pass_count > 0:
                record["readability_score"] = sum(evaluate_result) / pass_count

            record["quality_score"] = sum(evaluate_result) / count

        return pd.DataFrame(records)

    def score(self):
        records = self.detail_records()
        metrics = [
            "invalid_rate",
            "illegal rate",
            "pass_rate",
            "readability_score",
            "quality_score",
        ]
        score = {}
        for metric in metrics:
            score[metric] = records[metric].mean()

        for key in records.keys():
            if (
                key not in metrics
                and key != "id"
                and key != "chart"
                and key != "hardness"
            ):
                score[key] = records[key].mean()

        return score


def convert_svg_to_base64(svg_string):
    png_string = cairosvg.svg2png(bytestring=svg_string)
    base64_encoded = base64.b64encode(png_string).decode("utf-8")
    return f"data:image/png;base64,{base64_encoded}"


class Evaluator:
    def __init__(self, webdriver_path=None, vision_model=None):
        self.webdriver_path = webdriver_path
        self.vision_model = vision_model

    def evaluate(self, agent, dataset, config):
        use_logs = False
        evaluation_details = []
        if "logs" in config:
            log_folder = config["logs"]
            isExists = os.path.exists(log_folder)
            try:
                if not isExists:
                    os.makedirs(log_folder)
                logging.basicConfig(
                    level=logging.INFO,
                    filename=log_folder / "evaluation.log",
                    filemode="a",
                    format="%(levelname)s: %(message)s",
                )
                use_logs = True
            except Exception as e:
                print(e)

        for instance in dataset.benchmark:
            codes = []
            instance_results = []
            nl_queries = instance["nl_queries"]
            tables = instance["tables"]

            if use_logs:
                instanceFolder = log_folder / instance["id"]
                path = instanceFolder / "result.json"
                if os.path.exists(path):
                    with open(path, "r") as f:
                        data = json.load(f)
                        if "codes" in data and "evaluations" in data:
                            instance_results = []
                            for query_result in data["evaluations"]:
                                results = [
                                    CheckResult(
                                        answer=result["answer"],
                                        aspect=result["aspect"],
                                        rationale=result["rationale"],
                                    )
                                    for result in query_result
                                ]
                                instance_results.append(results)
                            evaluation_details.append(
                                EvaluationDetail(instance["id"], instance_results)
                            )
                            continue
                else:
                    logging.info(f"Instance ({instance['id']}) evaluation began.")
                    isExists = os.path.exists(instanceFolder)
                    if not isExists:
                        os.makedirs(instanceFolder)

            for index in range(len(nl_queries)):
                nl_query = nl_queries[index]
                if index < len(codes):
                    code = codes[index]
                    context = {}
                    context["tables"] = tables
                else:
                    code, context = agent.generate(nl_query, tables, config)
                    codes.append(code)
                if code is None:
                    results = [
                        CheckResult(
                            answer=False,
                            aspect="generation",
                            rationale="Code generation failed.",
                        )
                    ]
                else:
                    context["library"] = config["library"]
                    if use_logs:
                        results = self.validity_check(
                            code, context, agent, instanceFolder / f"{index}.svg"
                        )
                    else:
                        results = self.validity_check(code, context, agent)

                    pass_validity = all([result.answer for result in results])
                    if pass_validity:
                        ground_truth = {
                            "chart": instance["chart"],
                            "vis_obj": instance["vis_obj"],
                            "meta_info": instance["query_meta"][index],
                        }
                        results += self.legality_check(context, ground_truth)

                    pass_legality = all([result.answer for result in results])
                    if pass_legality:
                        results += self.readability_evaluate(context, nl_query)

                instance_results.append(results)

            evaluation_details.append(
                EvaluationDetail(instance["id"], instance_results)
            )
            if use_logs:
                logging.info(f"Instance ({instance['id']}) evaluation finished.")
                # convert CheckResult to json
                instance_results = [
                    [result.get_json() for result in results]
                    for results in instance_results
                ]
                with open(log_folder / (instance["id"] + "/result.json"), "w") as f:
                    f.write(
                        json.dumps({"codes": codes, "evaluations": instance_results})
                    )
                    f.close()
        return EvaluationResult(dataset, evaluation_details)

    def execute(self, code, context, agent, log_name=None) -> CheckResult:
        result = agent.execute(code, context, log_name)
        if result.status is False:
            return CheckResult(
                answer=False, aspect="code execution", rationale=result.error_msg
            )

        context["svg_string"] = result.svg_string
        return CheckResult(
            answer=True,
            aspect="code execution",
            rationale="Code executed successfully.",
        )

    def surface_form_check(self, code) -> CheckResult:
        if "plt.show()" not in code:
            return CheckResult(
                answer=False,
                aspect="surface-form check",
                rationale="Did not plot visualization.",
            )
        else:
            return CheckResult(
                answer=True,
                aspect="surface-form check",
                rationale="Plotted visualization.",
            )

    def validity_check(self, code, context, agent, log_name=None) -> list[CheckResult]:
        results = []
        result = self.execute(code, context, agent, log_name)
        results.append(result)
        if result.answer:
            result = self.surface_form_check(code)
            results.append(result)

        return results

    def deconstruction(self, context) -> CheckResult:
        svg_string = context["svg_string"]
        library = context["library"]
        if library == "seaborn":
            library = "matplotlib"
        try:
            chart_info, msg = deconstruct(svg_string, library)
            if chart_info is None:
                return CheckResult(
                    answer=False,
                    aspect="deconstruction",
                    rationale=msg,
                )
            context.update(chart_info)
            return CheckResult(
                answer=True,
                aspect="deconstruction",
                rationale="Deconstructed the chart successfully.",
            )
        except:
            return CheckResult(
                answer=False,
                aspect="deconstruction",
                rationale="Cannot parse the visualization.",
            )

    def chart_type_check(self, context, ground_truth) -> CheckResult:
        answer, rationale = chart_check(
            context,
            ground_truth["chart"],
            (
                ground_truth["meta_info"]["stacked_bar"]
                if "stacked_bar" in ground_truth["meta_info"]
                else None
            ),
        )
        return CheckResult(
            answer=answer,
            aspect="chart type check",
            rationale=rationale,
        )

    def data_check(self, context, ground_truth) -> CheckResult:
        answer, rationale = data_check(
            context,
            ground_truth["vis_obj"],
            ground_truth["meta_info"]["channel_specified"],
        )
        return CheckResult(
            answer=answer,
            aspect="data check",
            rationale=rationale,
        )

    def order_check(self, context, ground_truth) -> CheckResult:
        answer, rationale = order_check(
            context,
            ground_truth["vis_obj"],
            (
                ground_truth["meta_info"]["sort_by"]
                if "sort_by" in ground_truth["meta_info"]
                else None
            ),
        )
        return CheckResult(
            answer=answer,
            aspect="order check",
            rationale=rationale,
        )

    def legality_check(self, context, ground_truth) -> list[CheckResult]:
        results = []
        result = self.deconstruction(context)
        results.append(result)
        if result.answer:
            chart_type_check_result = self.chart_type_check(context, ground_truth)
            data_check_result = self.data_check(context, ground_truth)
            results.append(chart_type_check_result)
            results.append(data_check_result)
            if data_check_result.answer and ground_truth["vis_obj"]["sort"] is not None:
                self.order_check(context, ground_truth)
                results.append(self.order_check(context, ground_truth))

        return results

    def layout_check(self, context) -> CheckResult:
        assert "svg_string" in context
        assert self.webdriver_path is not None

        answer, rationale = layout_check(context, self.webdriver_path)
        return CheckResult(
            answer=answer,
            aspect="layout check",
            rationale=rationale,
        )

    def scale_and_ticks_check(self, context, query) -> CheckResult:
        assert "base64" in context and "encoding" in context and "chart" in context
        assert self.vision_model is not None

        answer, rationale = scale_and_ticks_check(context, query, self.vision_model)
        return CheckResult(
            answer=answer,
            aspect="scale and ticks check",
            rationale=rationale,
        )

    def readability_evaluate(self, context, query: str) -> list[CheckResult]:
        results = []
        if self.webdriver_path:
            layout_result = self.layout_check(context)
            if layout_result.answer is not None:
                results.append(layout_result)

        if self.vision_model:
            context["base64"] = convert_svg_to_base64(context["svg_string"])
            scale_and_ticks_result = self.scale_and_ticks_check(context, query)
            if scale_and_ticks_result.answer is not None:
                results.append(scale_and_ticks_result)

            aspect_format = {
                "layout check": "Overflow/Overlap",
                "scale and ticks check": "Scale/Ticks",
            }
            reviews = [
                {
                    "aspect": aspect_format[result.aspect],
                    "content": result.rationale,
                }
                for result in results
            ]
            context["reviews"] = reviews

            answer, rationale = readability_check(context, query, self.vision_model)
            if answer is not None:
                readability_result = CheckResult(
                    answer=answer,
                    aspect="readability check",
                    rationale=rationale,
                )
                results.append(readability_result)

        return results



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

from datetime import datetime

from dateutil import parser

TIME_MAP = {
    "mon": "monday",
    "tue": "tuesday",
    "wed": "wednesday",
    "thu": "thursday",
    "fri": "friday",
    "sat": "saturday",
    "sun": "sunday",
    "jan": "january",
    "feb": "february",
    "mar": "march",
    "apr": "april",
    "may": "may",
    "jun": "june",
    "jul": "july",
    "aug": "august",
    "sep": "september",
    "sept": "september",
    "oct": "october",
    "nov": "november",
    "dec": "december",
    "mon": "monday",
    "tue": "tuesday",
    "wed": "wednesday",
    "thu": "thursday",
    "thur": "thursday",
    "fri": "friday",
    "sat": "saturday",
    "sun": "sunday",
}
WEEKDAYS = ["mon", "tue", "wed", "thu", "fri", "sat", "sun"]
MONTHS = [
    "jan",
    "feb",
    "mar",
    "apr",
    "may",
    "jun",
    "jul",
    "aug",
    "sep",
    "oct",
    "nov",
    "dec",
]


def is_month_or_weekday(s: str):
    if isinstance(s, str):
        if s.lower() in TIME_MAP or (
            s[0:3].lower() in TIME_MAP and s.lower() == TIME_MAP[s[0:3].lower()]
        ):
            return True
    return False


def convert_month_or_weekday_to_int(s: str) -> int:
    if is_month_or_weekday(s):
        if s[0:3].lower() in WEEKDAYS:
            return WEEKDAYS.index(s[0:3].lower()) + 1
        if s[0:3].lower() in MONTHS:
            return MONTHS.index(s[0:3].lower()) + 1
    return -1


def is_datetime(s):
    # consider month and weekday as nominal
    if is_month_or_weekday(s):
        return False
    try:
        parser.parse(s)
        return True
    except ValueError:
        return False


def check_time_format(time_str, time_format):
    try:
        datetime.strptime(time_str, time_format)
        return True
    except ValueError:
        return False


def parse_time_to_timestamp(time_str):
    # 0:00 is prone to bias
    if check_time_format(time_str, "%Y"):
        time_str = time_str + "-01-01 00:00:10"
    elif check_time_format(time_str, "%Y-%m"):
        time_str = time_str + "-01 00:00:10"
    elif check_time_format(time_str, "%Y-%m-%d"):
        time_str = time_str + " 00:00:10"

    try:
        parsed_time = parser.parse(time_str)
        timestamp = parsed_time.timestamp()
        return timestamp
    except Exception:
        return None


def parse_timestamp_to_time(timestamp):
    # todo: extract date format
    date_format = "%Y-%m-%d"
    try:
        parsed_time = datetime.fromtimestamp(timestamp)
        time_str = parsed_time.strftime(date_format)
        return time_str
    except Exception:
        return None


# handle case like 2008.436089
def parse_number_to_time(number):
    if number > 0 and number < 2999:
        timestamp = parse_time_to_timestamp(str(int(number)))
        timestamp += (number - int(number)) * 365 * 24 * 60 * 60
        return timestamp
    return number


def compare_time_strings(time_str1: str, time_str2: str):
    try:
        if parser.parse(time_str1).timestamp() == parser.parse(time_str2).timestamp():
            return True
    except Exception:
        pass

    try:
        str1 = TIME_MAP.get(time_str1.lower(), time_str1)
        str2 = TIME_MAP.get(time_str2.lower(), time_str2)

        if str1.lower() == str2.lower():
            return True

        if (
            is_month_or_weekday(str1)
            and str(convert_month_or_weekday_to_int(str1)) == str2
        ):
            return True
        if (
            is_month_or_weekday(str2)
            and str(convert_month_or_weekday_to_int(str2)) == str1
        ):
            return True

        if (
            is_month_or_weekday(str1)
            and str(convert_month_or_weekday_to_int(str1)) == str2
        ):
            return True
        if (
            is_month_or_weekday(str2)
            and str(convert_month_or_weekday_to_int(str2)) == str1
        ):
            return True

        if is_month_or_weekday(str2) and parse_time_to_timestamp(time_str1) is not None:
            # weekday
            if (
                str2.lower()
                == datetime.fromtimestamp(parse_time_to_timestamp(time_str1))
                .strftime("%A")
                .lower()
            ):
                return True
            # month
            if (
                str2.lower()
                == datetime.fromtimestamp(parse_time_to_timestamp(time_str1))
                .strftime("%B")
                .lower()
            ):
                return True
        if is_month_or_weekday(str1) and parse_time_to_timestamp(time_str2) is not None:
            # weekday
            if (
                str1.lower()
                == datetime.fromtimestamp(parse_time_to_timestamp(time_str2))
                .strftime("%A")
                .lower()
            ):
                return True
            # month
            if (
                str1.lower()
                == datetime.fromtimestamp(parse_time_to_timestamp(time_str2))
                .strftime("%B")
                .lower()
            ):
                return True

        return False
    except Exception:
        return False
